# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import contextlib
import os
import textwrap
from logging import getLogger
from stat import S_ISDIR, S_ISREG
from typing import Optional

from etl_entities.instance import Host

try:
    from pydantic.v1 import FilePath, SecretStr
except (ImportError, AttributeError):
    from pydantic import FilePath, SecretStr  # type: ignore[no-redef, assignment]

from onetl.connection.file_connection.file_connection import FileConnection
from onetl.connection.file_connection.mixins.rename_dir_mixin import RenameDirMixin
from onetl.hooks import slot, support_hooks
from onetl.impl import LocalPath, RemotePath

try:
    from paramiko import ProxyCommand, SSHClient, SSHConfig, WarningPolicy
    from paramiko.sftp_attr import SFTPAttributes
    from paramiko.sftp_client import SFTPClient
    from paramiko.ssh_exception import ConfigParseError
except (ImportError, NameError) as e:
    raise ImportError(
        textwrap.dedent(
            """
            Cannot import module "paramiko".

            Since onETL v0.7.0 you should install package as follows:
                pip install onetl[sftp]

            or
                pip install onetl[files]
            """,
        ).strip(),
    ) from e

SSH_CONFIG_PATH = LocalPath("~/.ssh/config").expanduser().resolve()

log = getLogger(__name__)


@support_hooks
class SFTP(FileConnection, RenameDirMixin):
    """SFTP file connection. |support_hooks|

    Based on `Paramiko library <https://pypi.org/project/paramiko/>`_.

    .. warning::

        Since onETL v0.7.0 to use SFTP connector you should install package as follows:

        .. code:: bash

            pip install onetl[s3]

            # or
            pip install onetl[files]

        See :ref:`install-files` installation instruction for more details.

    .. versionadded:: 0.1.0

    Parameters
    ----------
    host : str
        Host of SFTP source. For example: ``192.168.1.19``

    port : int, default: ``22``
        Port of SFTP source

    user : str
        User, which have access to the file source. For example: ``someuser``

    password : str, default: ``None``
        Password for file source connection

    key_file : str, default: ``None``
        the filename of optional private key(s) and/or certs to try for authentication

    timeout : int, default: ``10``
        How long to wait for the server to send data before giving up

    host_key_check : bool, default: ``False``
        set to True to enable searching for discoverable private key files in ``~/.ssh/``

    compress : bool, default: ``True``
        Set to True to turn on compression

    Examples
    --------

    SFTP file connection initialization

    .. code:: python

        from onetl.connection import SFTP

        sftp = SFTP(
            host="192.168.1.19",
            user="someuser",
            password="*****",
        )
    """

    host: Host
    port: int = 22
    user: Optional[str] = None
    password: Optional[SecretStr] = None
    key_file: Optional[FilePath] = None
    timeout: int = 10
    host_key_check: bool = False
    compress: bool = True

    @property
    def instance_url(self) -> str:
        return f"{self.__class__.__name__.lower()}://{self.host}:{self.port}"

    def __str__(self):
        return f"{self.__class__.__name__}[{self.host}:{self.port}]"

    @slot
    def path_exists(self, path: os.PathLike | str) -> bool:
        try:
            self.client.stat(os.fspath(path))
            return True
        except FileNotFoundError:
            return False

    def _get_client(self) -> SFTPClient:
        host_proxy, key_file = self._parse_user_ssh_config()

        client = SSHClient()
        client.load_system_host_keys()
        if not self.host_key_check:
            # Default is RejectPolicy
            client.set_missing_host_key_policy(WarningPolicy())  # noqa: S507

        client.connect(
            hostname=self.host,
            port=self.port,
            username=self.user,
            password=self.password.get_secret_value() if self.password else None,
            key_filename=key_file,
            timeout=self.timeout,
            compress=self.compress,
            sock=host_proxy,
        )

        return client.open_sftp()

    def _is_client_closed(self, client: SFTPClient) -> bool:
        return not client.sock or client.sock.closed

    def _close_client(self, client: SFTPClient) -> None:
        client.close()

    def _parse_user_ssh_config(self) -> tuple[str | None, str | None]:
        host_proxy = None
        key_file = os.fspath(self.key_file) if self.key_file else None

        if not SSH_CONFIG_PATH.exists() or not SSH_CONFIG_PATH.is_file():
            return host_proxy, key_file

        try:
            ssh_conf = SSHConfig()
            ssh_conf.parse(SSH_CONFIG_PATH.read_text())
            host_info = ssh_conf.lookup(self.host) or {}
            if host_info.get("proxycommand"):
                host_proxy = ProxyCommand(host_info.get("proxycommand"))

            if not (self.password or key_file) and host_info.get("identityfile"):
                key_file = host_info.get("identityfile")[0]
        except ConfigParseError:
            log.exception("Failed to parse SSH config")

        return host_proxy, key_file

    def _create_dir(self, path: RemotePath) -> None:
        try:
            self.client.stat(os.fspath(path))
        except Exception:
            for parent in reversed(path.parents):
                try:  # noqa: WPS505
                    self.client.stat(os.fspath(parent))
                except Exception:
                    self.client.mkdir(os.fspath(parent))

            self.client.mkdir(os.fspath(path))

    def _upload_file(self, local_file_path: RemotePath, remote_file_path: RemotePath) -> None:
        self.client.put(os.fspath(local_file_path), os.fspath(remote_file_path))

    def _rename_file(self, source: RemotePath, target: RemotePath) -> None:
        with contextlib.suppress(OSError):
            self.client.posix_rename(os.fspath(source), os.fspath(target))
            return

        # posix rename extension is not supported by server
        # if OSError was caused by permissions error, client.rename will raise this exception again
        self.client.rename(os.fspath(source), os.fspath(target))

    _rename_dir = _rename_file

    def _download_file(self, remote_file_path: RemotePath, local_file_path: RemotePath) -> None:
        self.client.get(os.fspath(remote_file_path), os.fspath(local_file_path))

    def _remove_dir(self, path: RemotePath) -> None:
        self.client.rmdir(os.fspath(path))

    def _remove_file(self, remote_file_path: RemotePath) -> None:
        self.client.remove(os.fspath(remote_file_path))

    def _scan_entries(self, path: RemotePath) -> list[SFTPAttributes]:
        return self.client.listdir_attr(os.fspath(path))

    def _is_dir(self, path: RemotePath) -> bool:
        stat: SFTPAttributes = self.client.stat(os.fspath(path))
        return S_ISDIR(stat.st_mode)

    def _is_file(self, path: RemotePath) -> bool:
        stat: SFTPAttributes = self.client.stat(os.fspath(path))
        return S_ISREG(stat.st_mode)

    def _get_stat(self, path: RemotePath) -> SFTPAttributes:
        # underlying SFTP client already return `os.stat_result`-like class
        return self.client.stat(os.fspath(path))

    def _extract_name_from_entry(self, entry: SFTPAttributes) -> str:
        return entry.filename

    def _is_dir_entry(self, top: RemotePath, entry: SFTPAttributes) -> bool:
        return S_ISDIR(entry.st_mode)

    def _is_file_entry(self, top: RemotePath, entry: SFTPAttributes) -> bool:
        return S_ISREG(entry.st_mode)

    def _extract_stat_from_entry(self, top: RemotePath, entry: SFTPAttributes) -> SFTPAttributes:
        return entry

    def _read_text(self, path: RemotePath, encoding: str, **kwargs) -> str:
        with self.client.open(os.fspath(path), mode="r", **kwargs) as file:
            return file.read().decode(encoding)

    def _read_bytes(self, path: RemotePath, **kwargs) -> bytes:
        with self.client.open(os.fspath(path), mode="r", **kwargs) as file:
            return file.read()

    def _write_text(self, path: RemotePath, content: str, encoding: str, **kwargs) -> None:
        with self.client.open(os.fspath(path), mode="w", **kwargs) as file:
            file.write(content.encode(encoding))

    def _write_bytes(self, path: RemotePath, content: bytes, **kwargs) -> None:
        with self.client.open(os.fspath(path), mode="w", **kwargs) as file:
            file.write(content)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional

try:
    from pydantic.v1 import Field, SecretStr, validator
except (ImportError, AttributeError):
    from pydantic import Field, SecretStr, validator  # type: ignore[no-redef, assignment]

from onetl._util.file import is_file_readable
from onetl._util.spark import stringify
from onetl.impl import GenericOptions, LocalPath

if TYPE_CHECKING:
    from onetl.connection import Kafka

from onetl.connection.db_connection.kafka.kafka_protocol import KafkaProtocol


class KafkaSSLProtocol(KafkaProtocol, GenericOptions):
    """
    Connect to Kafka using ``SSL`` or ``SASL_SSL`` security protocols.

    For more details see:

    * `Kafka Documentation <https://kafka.apache.org/documentation/#producerconfigs_ssl.keystore.location>`_
    * `IBM Documentation <https://www.ibm.com/docs/en/cloud-paks/cp-biz-automation/19.0.x?topic=fcee-kafka-using-ssl-kerberos-authentication>`_
    * `How to use PEM Certificates with Kafka <https://codingharbour.com/apache-kafka/using-pem-certificates-with-apache-kafka/>`_

    .. versionadded:: 0.9.0

    Examples
    --------

    Pass PEM key and certificates as files located on Spark driver host:

    .. code:: python

        from pathlib import Path

        # Just read existing files located on host, and pass key and certificates as strings
        protocol = Kafka.SSLProtocol(
            keystore_type="PEM",
            keystore_certificate_chain=Path("path/to/user.crt").read_text(),
            keystore_key=Path("path/to/user.key").read_text(),
            truststore_type="PEM",
            truststore_certificates=Path("/path/to/server.crt").read_text(),
        )

    Pass PEM key and certificates as raw strings:

    .. code:: python

        protocol = Kafka.SSLProtocol(
            keystore_type="PEM",
            keystore_certificate_chain="-----BEGIN CERTIFICATE-----\\nMIIDZjC...\\n-----END CERTIFICATE-----",
            keystore_key="-----BEGIN PRIVATE KEY-----\\nMIIEvg..\\n-----END PRIVATE KEY-----",
            truststore_type="PEM",
            truststore_certificates="-----BEGIN CERTIFICATE-----\\nMICC...\\n-----END CERTIFICATE-----",
        )

    Pass custom options:

    .. code:: python

        protocol = Kafka.SSLProtocol.parse(
            {
                # Just the same options as above, but using Kafka config naming with dots
                "ssl.keystore.type": "PEM",
                "ssl.keystore.certificate_chain": "-----BEGIN CERTIFICATE-----\\nMIIDZjC...\\n-----END CERTIFICATE-----",
                "ssl.keystore.key": "-----BEGIN PRIVATE KEY-----\\nMIIEvg..\\n-----END PRIVATE KEY-----",
                "ssl.truststore.type": "PEM",
                "ssl.truststore.certificates": "-----BEGIN CERTIFICATE-----\\nMICC...\\n-----END CERTIFICATE-----",
                # Any option starting from "ssl." is passed to Kafka client as-is
                "ssl.protocol": "TLSv1.3",
            }
        )

    .. dropdown :: Not recommended

        These options are error-prone and have several drawbacks, so it is not recommended to use them.

        Passing PEM certificates as files:

        * ENCRYPT ``user.key`` file with password ``"some password"`` `using PKCS#8 scheme <https://www.mkssoftware.com/docs/man1/openssl_pkcs8.1.asp>`_.
        * Save encrypted key to file ``/path/to/user/encrypted_key_with_certificate_chain.pem``.
        * Then append user certificate to the end of this file.
        * Deploy this file (and server certificate too) to **EVERY** host Spark could run (both driver and executors).
        * Then pass file locations and password for key decryption to options below.

        .. code:: python

            protocol = Kafka.SSLProtocol(
                keystore_type="PEM",
                keystore_location="/path/to/user/encrypted_key_with_certificate_chain.pem",
                key_password="some password",
                truststore_type="PEM",
                truststore_location="/path/to/server.crt",
            )

        Passing JKS (Java Key Store) location:

        * `Add user key and certificate to JKS keystore <https://stackoverflow.com/a/4326346>`_.
        * `Add server certificate to JKS truststore <https://stackoverflow.com/a/373307>`_.
        * This should be done on **EVERY** host Spark could run (both driver and executors).
        * Pass keystore and truststore paths to options below, as well as passwords for accessing these stores:

        .. code:: python

            protocol = Kafka.SSLProtocol(
                keystore_type="JKS",
                keystore_location="/usr/lib/jvm/default/lib/security/keystore.jks",
                keystore_password="changeit",
                truststore_type="JKS",
                truststore_location="/usr/lib/jvm/default/lib/security/truststore.jks",
                truststore_password="changeit",
            )
    """

    keystore_type: str = Field(alias="ssl.keystore.type")
    keystore_location: Optional[LocalPath] = Field(default=None, alias="ssl.keystore.location")
    keystore_password: Optional[SecretStr] = Field(default=None, alias="ssl.keystore.password")
    keystore_certificate_chain: Optional[str] = Field(default=None, alias="ssl.keystore.certificate.chain", repr=False)
    keystore_key: Optional[SecretStr] = Field(default=None, alias="ssl.keystore.key")
    # https://knowledge.informatica.com/s/article/145442?language=en_US
    key_password: Optional[SecretStr] = Field(default=None, alias="ssl.key.password")
    truststore_type: str = Field(alias="ssl.truststore.type")
    truststore_location: Optional[LocalPath] = Field(default=None, alias="ssl.truststore.location")
    truststore_password: Optional[SecretStr] = Field(default=None, alias="ssl.truststore.password")
    truststore_certificates: Optional[str] = Field(default=None, alias="ssl.truststore.certificates", repr=False)

    class Config:
        known_options = {"ssl.*"}
        strip_prefixes = ["kafka."]
        extra = "allow"

    def get_options(self, kafka: Kafka) -> dict:
        result = self.dict(by_alias=True, exclude_none=True)
        if kafka.auth:
            result["security.protocol"] = "SASL_SSL"
        else:
            result["security.protocol"] = "SSL"
        return stringify(result)

    def cleanup(self, kafka: Kafka) -> None:
        # nothing to cleanup
        pass

    @validator("keystore_location", "truststore_location")
    def validate_path(cls, value: LocalPath) -> Path:
        return is_file_readable(value)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing_extensions import Protocol, runtime_checkable

from onetl.base.path_stat_protocol import PathStatProtocol
from onetl.base.pure_path_protocol import PurePathProtocol


@runtime_checkable
class PathProtocol(PurePathProtocol, Protocol):
    """
    Generic protocol for :obj:`pathlib.Path` like objects.

    Includes only minimal set of methods which allow to determine path type (file, directory) and existence
    """

    def is_dir(self) -> bool:
        """
        Checks if this path is a directory
        """

    def is_file(self) -> bool:
        """
        Checks if this path is a file
        """

    def exists(self) -> bool:
        """
        Checks if this path exists
        """


@runtime_checkable
class PathWithStatsProtocol(PathProtocol, Protocol):
    """
    Protocol for ``pathlib.Path``-like file objects.

    Includes only minimal set of methods which allow to determine if file exists, or get stats, e.g. size
    """

    def stat(self) -> PathStatProtocol:
        """
        Returns stats object with file information
        """



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from abc import ABC, abstractmethod

from onetl.base.path_protocol import PathProtocol


class BaseFileFilter(ABC):
    """
    Base file filter interface.

    Filters used by several onETL components, including :ref:`file-downloader` and :ref:`file-mover`,
    to determine if a file should be handled or not.

    All filters are stateless.

    .. versionadded:: 0.8.0
    """

    @abstractmethod
    def match(self, path: PathProtocol) -> bool:
        """
        Returns ``True`` if path is matching the filter, ``False`` otherwise

        .. versionadded:: 0.8.0

        Examples
        --------

        from onetl.impl import LocalPath

        >>> filter.match(LocalPath("/path/to/file.csv"))
        True
        >>> filter.match(LocalPath("/path/to/excluded.csv"))
        False
        >>> filter.match(LocalPath("/path/to/file.csv"))
        True
        """



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, ContextManager

if TYPE_CHECKING:
    from pyspark.sql import DataFrameReader, DataFrameWriter, SparkSession


class BaseReadableFileFormat(ABC):
    """
    Representation of readable file format.

    .. versionadded:: 0.9.0
    """

    @abstractmethod
    def check_if_supported(self, spark: SparkSession) -> None:
        """
        Check if Spark session does support this file format. |support_hooks|

        .. versionadded:: 0.9.0

        Raises
        ------
        RuntimeError
            If file format is not supported.
        """

    @abstractmethod
    def apply_to_reader(self, reader: DataFrameReader) -> DataFrameReader | ContextManager[DataFrameReader]:
        """
        Apply provided format to :obj:`pyspark.sql.DataFrameReader`. |support_hooks|

        .. versionadded:: 0.9.0

        Returns
        -------
        :obj:`pyspark.sql.DataFrameReader`
            DataFrameReader with options applied.

        ``ContextManager[DataFrameReader]``
            If returned context manager, it will be entered before reading data and exited after creating a DataFrame.
            Context manager's ``__enter__`` method should return :obj:`pyspark.sql.DataFrameReader` instance.
        """


class BaseWritableFileFormat(ABC):
    """
    Representation of writable file format.

    .. versionadded:: 0.9.0
    """

    @abstractmethod
    def check_if_supported(self, spark: SparkSession) -> None:
        """
        Check if Spark session does support this file format. |support_hooks|

        .. versionadded:: 0.9.0

        Raises
        ------
        RuntimeError
            If file format is not supported.
        """

    @abstractmethod
    def apply_to_writer(self, writer: DataFrameWriter) -> DataFrameWriter | ContextManager[DataFrameWriter]:
        """
        Apply provided format to :obj:`pyspark.sql.DataFrameWriter`. |support_hooks|

        .. versionadded:: 0.9.0

        Returns
        -------
        :obj:`pyspark.sql.DataFrameWriter`
            DataFrameWriter with options applied.

        ``ContextManager[DataFrameWriter]``
            If returned context manager, it will be entered before writing and exited after writing a DataFrame.
            Context manager's ``__enter__`` method should return :obj:`pyspark.sql.DataFrameWriter` instance.
        """



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import ftplib  # noqa: S402  # nosec
import os
import textwrap
from logging import getLogger
from typing import Optional

from etl_entities.instance import Host

try:
    from pydantic.v1 import SecretStr
except (ImportError, AttributeError):
    from pydantic import SecretStr  # type: ignore[no-redef, assignment]

from onetl.base import PathStatProtocol
from onetl.connection.file_connection.file_connection import FileConnection
from onetl.connection.file_connection.mixins.rename_dir_mixin import RenameDirMixin
from onetl.hooks import slot, support_hooks
from onetl.impl import LocalPath, RemotePath
from onetl.impl.remote_path_stat import RemotePathStat

try:
    from ftputil import FTPHost
    from ftputil import session as ftp_session
except (ImportError, NameError) as e:
    raise ImportError(
        textwrap.dedent(
            """
            Cannot import module "ftputil".

            Since onETL v0.7.0 you should install package as follows:
                pip install onetl[ftp]

            or
                pip install onetl[files]
            """,
        ).strip(),
    ) from e

log = getLogger(__name__)


@support_hooks
class FTP(FileConnection, RenameDirMixin):
    """FTP file connection. |support_hooks|

    Based on `FTPUtil library <https://pypi.org/project/ftputil/>`_.

    .. warning::

        Since onETL v0.7.0 to use FTP connector you should install package as follows:

        .. code:: bash

            pip install onetl[ftp]

            # or
            pip install onetl[files]

        See :ref:`install-files` installation instruction for more details.

    .. versionadded:: 0.1.0

    Parameters
    ----------
    host : str
        Host of FTP source. For example: ``ftp.domain.com``

    port : int, default: ``21``
        Port of FTP source

    user : str, default: ``None``
        User, which have access to the file source. For example: ``someuser``.

        ``None`` means that the user is anonymous.

    password : str, default: ``None``
        Password for file source connection.

        ``None`` means that the user is anonymous.

    Examples
    --------

    FTP file connection initialization

    .. code:: python

        from onetl.connection import FTP

        ftp = FTP(
            host="ftp.domain.com",
            user="someuser",
            password="*****",
        )
    """

    host: Host
    port: int = 21
    user: Optional[str] = None
    password: Optional[SecretStr] = None

    @property
    def instance_url(self) -> str:
        return f"{self.__class__.__name__.lower()}://{self.host}:{self.port}"

    def __str__(self):
        return f"{self.__class__.__name__}[{self.host}:{self.port}]"

    @slot
    def path_exists(self, path: os.PathLike | str) -> bool:
        return self.client.path.exists(os.fspath(path))

    def _get_client(self) -> FTPHost:
        """
        Returns a FTP connection object
        """

        session_factory = ftp_session.session_factory(
            base_class=ftplib.FTP,
            port=self.port,
            encrypt_data_channel=True,
            debug_level=0,
        )

        return FTPHost(
            self.host,
            self.user,
            self.password.get_secret_value() if self.password else None,
            session_factory=session_factory,
        )

    def _is_client_closed(self, client: FTPHost) -> bool:
        return client.closed

    def _close_client(self, client: FTPHost) -> None:
        client.close()

    def _remove_dir(self, path: RemotePath) -> None:
        self.client.rmdir(os.fspath(path))

    def _upload_file(self, local_file_path: LocalPath, remote_file_path: RemotePath) -> None:
        self.client.upload(os.fspath(local_file_path), os.fspath(remote_file_path))

    def _rename_file(self, source: RemotePath, target: RemotePath) -> None:
        self.client.rename(os.fspath(source), os.fspath(target))

    _rename_dir = _rename_file

    def _download_file(self, remote_file_path: RemotePath, local_file_path: LocalPath) -> None:
        self.client.download(os.fspath(remote_file_path), os.fspath(local_file_path))

    def _remove_file(self, remote_file_path: RemotePath) -> None:
        self.client.remove(os.fspath(remote_file_path))

    def _create_dir(self, path: RemotePath) -> None:
        self.client.makedirs(os.fspath(path), exist_ok=True)

    def _scan_entries(self, path: RemotePath) -> list[str]:
        return self.client.listdir(os.fspath(path))

    def _is_dir(self, path: RemotePath) -> bool:
        return self.client.path.isdir(os.fspath(path))

    def _is_file(self, path: RemotePath) -> bool:
        return self.client.path.isfile(os.fspath(path))

    def _get_stat(self, path: RemotePath) -> PathStatProtocol:
        if path == RemotePath("/"):
            # FTP does not allow to call stat on root directory, do nothing
            return RemotePathStat()

        # underlying FTP client already return `os.stat_result`-like class`
        return self.client.stat(os.fspath(path))

    def _read_text(self, path: RemotePath, encoding: str, **kwargs) -> str:
        with self.client.open(os.fspath(path), mode="r", encoding=encoding, **kwargs) as file:
            return file.read()

    def _read_bytes(self, path: RemotePath, **kwargs) -> bytes:
        with self.client.open(os.fspath(path), mode="rb", **kwargs) as file:
            return file.read()

    def _write_text(self, path: RemotePath, content: str, encoding: str, **kwargs) -> None:
        with self.client.open(os.fspath(path), mode="w", encoding=encoding, **kwargs) as file:
            file.write(content)

    def _write_bytes(self, path: RemotePath, content: bytes, **kwargs) -> None:
        with self.client.open(os.fspath(path), mode="wb", **kwargs) as file:
            file.write(content)

    def _extract_name_from_entry(self, entry: str) -> str:
        return entry

    def _is_dir_entry(self, top: RemotePath, entry: str) -> bool:
        return self._is_dir(top / entry)

    def _is_file_entry(self, top: RemotePath, entry: str) -> bool:
        return self._is_file(top / entry)

    def _extract_stat_from_entry(self, top: RemotePath, entry: str) -> PathStatProtocol:
        return self._get_stat(top / entry)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from pyspark.sql.types import StructType


@runtime_checkable
class ContainsGetDFSchemaMethod(Protocol):
    """
    Protocol for objects containing ``get_df_schema`` method
    """

    def get_df_schema(
        self,
        source: str,
        columns: list[str] | None = None,
    ) -> StructType:
        """
        Description of the dataframe schema. |support_hooks|
        """
        ...



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING

from onetl.impl.frozen_model import FrozenModel

if TYPE_CHECKING:
    from onetl.connection import Kafka

from onetl.connection.db_connection.kafka.kafka_protocol import KafkaProtocol


class KafkaPlaintextProtocol(KafkaProtocol, FrozenModel):
    """
    Connect to Kafka using ``PLAINTEXT`` or ``SASL_PLAINTEXT`` security protocols.

    .. warning::

        Not recommended to use on production environments.
        Prefer :obj:`SSLProtocol <onetl.connection.db_connection.kafka.kafka_ssl_protocol.KafkaSSLProtocol>`.

    .. versionadded:: 0.9.0

    Examples
    --------

    .. code:: python

        # No options
        protocol = Kafka.PlaintextProtocol()
    """

    def get_options(self, kafka: Kafka) -> dict:
        # Access to Kafka is needed to select the type of protocol depending on the authentication scheme.
        if kafka.auth:
            return {"security.protocol": "SASL_PLAINTEXT"}
        return {"security.protocol": "PLAINTEXT"}

    def cleanup(self, kafka: Kafka) -> None:
        # nothing to cleanup
        pass



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
import textwrap

try:
    from pydantic.v1 import Field
except (ImportError, AttributeError):
    from pydantic import Field  # type: ignore[no-redef, assignment]

from onetl._metrics.driver import SparkDriverMetrics
from onetl._metrics.executor import SparkExecutorMetrics
from onetl._metrics.input import SparkInputMetrics
from onetl._metrics.output import SparkOutputMetrics
from onetl.impl import BaseModel

INDENT = " " * 4


class SparkCommandMetrics(BaseModel):
    input: SparkInputMetrics = Field(default_factory=SparkInputMetrics)
    output: SparkOutputMetrics = Field(default_factory=SparkOutputMetrics)
    driver: SparkDriverMetrics = Field(default_factory=SparkDriverMetrics)
    executor: SparkExecutorMetrics = Field(default_factory=SparkExecutorMetrics)

    @property
    def is_empty(self) -> bool:
        return all([self.input.is_empty, self.output.is_empty])

    def update(self, other: SparkCommandMetrics) -> SparkCommandMetrics:
        self.input.update(other.input)
        self.output.update(other.output)
        self.driver.update(other.driver)
        self.executor.update(other.executor)
        return self

    @property
    def details(self) -> str:
        if self.is_empty:
            return "No data"

        result = []
        if not self.input.is_empty:
            result.append(f"Input:{os.linesep}{textwrap.indent(self.input.details, INDENT)}")
        if not self.output.is_empty:
            result.append(f"Output:{os.linesep}{textwrap.indent(self.output.details, INDENT)}")
        if not self.driver.is_empty:
            result.append(f"Driver:{os.linesep}{textwrap.indent(self.driver.details, INDENT)}")
        if not self.executor.is_empty:
            result.append(f"Executor:{os.linesep}{textwrap.indent(self.executor.details, INDENT)}")

        return os.linesep.join(result)

    def __str__(self):
        return self.details



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass
from typing import TYPE_CHECKING

from onetl._util.java import get_java_gateway, start_callback_server

if TYPE_CHECKING:
    from pyspark.sql import SparkSession


@dataclass
class BaseSparkListener:
    """Base no-op SparkListener implementation.

    See `SparkListener <https://spark.apache.org/docs/3.5.2/api/java/org/apache/spark/scheduler/SparkListener.html>`_ interface.
    """

    spark: SparkSession

    def activate(self):
        start_callback_server(self.spark)

        # passing python listener object directly to addSparkListener or removeSparkListener leads to creating new java object each time.
        # But removeSparkListener call has effect only on the same Java object passed to removeSparkListener.
        # So we need to explicitly create Java object, and then pass it both calls.
        gateway = get_java_gateway(self.spark)
        java_list = gateway.jvm.java.util.ArrayList()
        java_list.append(self)
        self._java_listener = java_list[0]

        spark_context = self.spark.sparkContext._jsc.sc()  # noqa: WPS437
        spark_context.addSparkListener(self._java_listener)

    def deactivate(self):
        with suppress(Exception):
            spark_context = self.spark.sparkContext._jsc.sc()  # noqa: WPS437
            spark_context.removeSparkListener(self._java_listener)

        with suppress(Exception):
            del self._java_listener

    def __enter__(self):
        self.activate()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.deactivate()

    def __del__(self):  # noqa: WPS603
        # If current object is collected by GC, deactivate listener
        # and free bind Java object
        self.deactivate()

    def equals(self, other):
        # Java does not provide proper way to get object id for comparison,
        # so we compare string representation which should contain some form of id
        return other.toString() == self._java_listener.toString()

    def toString(self):
        return type(self).__qualname__ + "@" + hex(id(self))

    def hashCode(self):
        return hash(self)

    # no cover: start
    # method names are important for Java interface compatibility!
    def onApplicationEnd(self, application):
        pass

    def onApplicationStart(self, application):
        pass

    def onBlockManagerAdded(self, block_manager):
        pass

    def onBlockManagerRemoved(self, block_manager):
        pass

    def onBlockUpdated(self, block):
        pass

    def onEnvironmentUpdate(self, environment):
        pass

    def onExecutorAdded(self, executor):
        pass

    def onExecutorMetricsUpdate(self, executor):
        pass

    def onExecutorRemoved(self, executor):
        pass

    def onExecutorBlacklisted(self, event):
        pass

    def onExecutorBlacklistedForStage(self, event):
        pass

    def onExecutorExcluded(self, event):
        pass

    def onExecutorExcludedForStage(self, event):
        pass

    def onExecutorUnblacklisted(self, event):
        pass

    def onExecutorUnexcluded(self, event):
        pass

    def onJobStart(self, event):
        pass

    def onJobEnd(self, event):
        pass

    def onNodeBlacklisted(self, node):
        pass

    def onNodeBlacklistedForStage(self, stage):
        pass

    def onNodeExcluded(self, node):
        pass

    def onNodeExcludedForStage(self, node):
        pass

    def onNodeUnblacklisted(self, node):
        pass

    def onNodeUnexcluded(self, node):
        pass

    def onOtherEvent(self, event):
        pass

    def onResourceProfileAdded(self, resource_profile):
        pass

    def onSpeculativeTaskSubmitted(self, task):
        pass

    def onStageCompleted(self, event):
        pass

    def onStageExecutorMetrics(self, metrics):
        pass

    def onStageSubmitted(self, event):
        pass

    def onTaskEnd(self, event):
        pass

    def onTaskGettingResult(self, task):
        pass

    def onTaskStart(self, event):
        pass

    def onUnpersistRDD(self, rdd):
        pass

    def onUnschedulableTaskSetAdded(self, task_set):
        pass

    def onUnschedulableTaskSetRemoved(self, task_set):
        pass

    # no cover: stop
    class Java:
        implements = ["org.apache.spark.scheduler.SparkListenerInterface"]



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import os

from onetl.plugins import import_plugins
from onetl.version import __version__


def plugins_auto_import():
    """
    Automatically import all onETL plugins.

    Executed while onETL is being imported.

    See :ref:`plugins` documentation.
    """
    plugins_enabled = os.getenv("ONETL_PLUGINS_ENABLED", "true").lower() != "false"
    if not plugins_enabled:
        return

    plugins_whitelist = list(filter(None, os.getenv("ONETL_PLUGINS_WHITELIST", "").split(",")))
    plugins_blacklist = list(filter(None, os.getenv("ONETL_PLUGINS_BLACKLIST", "").split(",")))

    import_plugins("onetl.plugins", whitelist=plugins_whitelist, blacklist=plugins_blacklist)


plugins_auto_import()



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import io
import json
import logging
import sys
from contextlib import redirect_stdout
from enum import Enum
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Collection, Iterable

from etl_entities.hwm import HWM
from typing_extensions import deprecated

if TYPE_CHECKING:
    from pyspark.sql import DataFrame

onetl_log = logging.getLogger("onetl")
root_log = logging.getLogger()

HALF_SCREEN_SIZE = 45
BASE_LOG_INDENT = 8
LOG_FORMAT = "%(asctime)s [%(levelname)-8s] %(threadName)s: %(message)s"
CLIENT_MODULES = {"hdfs", "paramiko", "ftputil", "minio", "webdav3", "pyspark"}

DISABLED = 9999  # CRITICAL is 50, we need even higher to disable all logs
logging.addLevelName(DISABLED, "DISABLED")

NOTICE = 5  # DEBUG is 10, we need lower value for less verbose logs even on debug level
logging.addLevelName(NOTICE, "NOTICE")


@deprecated(
    "Deprecated in 0.5.0 and will be removed in 1.0.0. Use 'setup_logging' instead",
    category=UserWarning,
)
def setup_notebook_logging(level: int | str = logging.INFO) -> None:
    """Set up onETL logging.

    What this function does:
        * Adds stderr logging handler
        * Changes root logger format to ``2023-05-31 11:22:33.456 [INFO]: message``
        * Changes root logger level to ``level``
        * Changes onETL logger level to ``level``
        * Disables loggers of underlying client modules

    .. note::

        Should be used only in IDEs (like Jupyter notebooks or PyCharm),
        or scripts (ETL pipelines).

    .. warning::

        Should **NOT** be used in applications, you should set up logging settings manually,
        according to your framework documentation.

    .. deprecated:: 0.5.0
        Use :obj:`~setup_logging` instead

    Parameters
    ----------
    level : ``int`` or ``str``, default ``INFO``
        Log level for onETL module
    """

    setup_logging(level)


def setup_logging(level: int | str = logging.INFO, enable_clients: bool = False) -> None:
    """Set up onETL logging.

    What this function does:
        * Adds stderr logging handler
        * Changes root logger format to ``2023-05-31 11:22:33.456 [INFO] MainThread: message``
        * Changes root logger level to ``level``
        * Changes onETL logger level to ``level``
        * Sets up logging level of underlying client modules

    .. note::

        Should be used only in IDEs (like Jupyter notebooks or PyCharm),
        or scripts (ETL pipelines).

    .. versionchanged:: 0.5.0
        Renamed ``setup_notebook_logging`` → ``setup_logging``

    Parameters
    ----------
    level : ``int`` or ``str``, default ``INFO``
        Log level for onETL module

    enable_clients : ``bool``, default ``False``
        If ``True``, enable logging of underlying client modules.
        Otherwise, set client modules log level to ``DISABLED``.

        .. note::

            For ``level="DEBUG"`` it is recommended to use ``enable_clients=True``

        .. versionadded:: 0.9.0
    """

    logging.basicConfig(level=level)
    set_default_logging_format()

    onetl_log.setLevel(level)
    setup_clients_logging(level if enable_clients else DISABLED)


def setup_clients_logging(level: int | str = DISABLED) -> None:
    """Set logging of underlying client modules used by onETL.

    Affected modules:
        * ``ftputil``
        * ``hdfs``
        * ``minio``
        * ``paramiko``
        * ``py4j``
        * ``pyspark``
        * ``webdav3``

    .. note::

        Can be used in applications, but it is recommended to set up these loggers
        according to your framework documentation.

    .. versionchanged:: 0.9.0
        Renamed ``disable_clients_logging`` → ``setup_clients_logging``

    Parameters
    ----------
    level : ``int`` or ``str``, default ``DISABLED``
        Log level for client modules

        .. note::

            For ``py4j``, logging level with maximum verbosity is ``INFO`` because ``DEBUG`` logs are
            totally unreadable.

        .. versionadded:: 0.9.0
    """

    for client_module in CLIENT_MODULES:
        logging.getLogger(client_module).setLevel(level)

    if isinstance(level, str):
        level = int(logging.getLevelName(level))
    logging.getLogger("py4j").setLevel(max(level, logging.INFO))


def set_default_logging_format() -> None:
    """Sets default logging format to preferred by onETL.

    Example log message: ``2023-05-31 11:22:33.456 [INFO] MainThread: message``

    .. note::

        Should be used only in IDEs (like Jupyter notebooks or PyCharm),
        or scripts (ETL pipelines).

    .. warning::

        Should **NOT** be used in applications, you should set up logging settings manually,
        according to your framework documentation.
    """

    handlers = onetl_log.handlers or root_log.handlers
    for handler in handlers:
        handler.setFormatter(logging.Formatter(LOG_FORMAT))


def _log(logger: logging.Logger, msg: str, *args, level: int = logging.INFO, stacklevel: int = 1, **kwargs) -> None:
    if sys.version_info >= (3, 8):
        # https://github.com/python/cpython/pull/7424
        logger.log(level, msg, *args, stacklevel=stacklevel + 1, **kwargs)  # noqa: WPS204
    else:
        logger.log(level, msg, *args, **kwargs)


def log_with_indent(
    logger: logging.Logger,
    inp: str,
    *args,
    indent: int = 0,
    level: int = logging.INFO,
    stacklevel: int = 1,
    **kwargs,
) -> None:
    """Log a message with indentation.

    Supports all positional and keyword arguments which ``logging.log`` support.

    Example
    -------

    .. code:: python

        log_with_indent(logger, "message")
        log_with_indent(
            logger,
            "message with additional %s",
            "indent",
            indent=4,
            level=logging.DEBUG,
        )

    .. code-block:: text

        INFO  onetl.module        message
        DEBUG onetl.module            message with additional indent

    """
    _log(logger, "%s" + inp, " " * (BASE_LOG_INDENT + indent), *args, level=level, stacklevel=stacklevel + 1, **kwargs)


def log_lines(
    logger: logging.Logger,
    inp: str,
    name: str | None = None,
    indent: int = 0,
    level: int = logging.INFO,
    stacklevel: int = 1,
):
    r"""Log input multiline string with indentation.

    Any input indentation is being stripped.

    Does NOT support variable substitution.

    Examples
    --------

    .. code:: python

        log_lines(logger, "line1\nline2")
        log_lines(logger, "  line1\n      line2\n  line3", level=logging.DEBUG)

    .. code-block:: text

        INFO  onetl.module        line1
        INFO  onetl.module        line2

        DEBUG onetl.module        line1
        DEBUG onetl.module            line2
        DEBUG onetl.module        line3

    """

    base_indent = " " * (BASE_LOG_INDENT + indent)
    stacklevel += 1
    for index, line in enumerate(dedent(inp).splitlines()):
        if name and not index:
            _log(logger, "%s%s = %s", base_indent, name, line, level=level, stacklevel=stacklevel)
        else:
            _log(logger, "%s%s", base_indent, line, level=level, stacklevel=stacklevel)


def log_json(
    logger: logging.Logger,
    inp: Any,
    name: str | None = None,
    indent: int = 0,
    level: int = logging.INFO,
    stacklevel: int = 1,
):
    """Log input object in JSON notation.

    Does NOT support variable substitution.

    Examples
    --------

    .. code:: python

        log_json(logger, ["item1", {"item2": "value2"}, None])

        log_json(logger, {"item2": "value2"}, name="myvar", level=logging.DEBUG)

    .. code-block:: text

        INFO  onetl.module        [
        INFO  onetl.module            "item1",
        INFO  onetl.module            {"item2": "value2"},
        INFO  onetl.module            null
        INFO  onetl.module        ]

        DEBUG onetl.module        myvar = {
        DEBUG onetl.module            "item2": "value2"
        DEBUG onetl.module        }

    """

    log_lines(logger, json.dumps(inp, indent=4), name, indent, level, stacklevel=stacklevel + 1)


def log_collection(
    logger: logging.Logger,
    name: str,
    collection: Iterable,
    max_items: int | None = None,
    indent: int = 0,
    level: int = logging.INFO,
    stacklevel: int = 1,
):
    """Log input collection.

    Does NOT support variable substitution.

    Examples
    --------

    .. code:: python

        log_collection(logger, "myvar", [])
        log_collection(logger, "myvar", ["item1", {"item2": "value2"}, None])
        log_collection(logger, "myvar", ["item1", "item2", "item3"], max_items=1)
        log_collection(
            logger,
            "myvar",
            ["item1", "item2", "item3"],
            max_items=1,
            level=logging.DEBUG,
        )

    .. code-block:: text

        INFO  onetl.module        myvar = []

        INFO  onetl.module        myvar = [
        INFO  onetl.module            'item1',
        INFO  onetl.module            {'item2': 'value2'},
        INFO  onetl.module            None,
        INFO  onetl.module        ]

        INFO onetl.module        myvar = [
        INFO onetl.module            'item1',
        INFO onetl.module            # ... 2 more items of type <class 'str'>
        INFO onetl.module            # change level to 'DEBUG' to print all values
        INFO onetl.module        ]

        DEBUG onetl.module        myvar = [
        DEBUG onetl.module            'item1',
        DEBUG onetl.module            'item2',
        DEBUG onetl.module            'item3',
        DEBUG onetl.module        ]

    """

    base_indent = " " * (BASE_LOG_INDENT + indent)
    stacklevel += 1
    items = list(collection)  # force convert all iterators to list to know size
    if not items:
        _log(logger, "%s%s = []", base_indent, name, level=level, stacklevel=stacklevel)
        return

    nested_indent = " " * (BASE_LOG_INDENT + indent + 4)
    _log(logger, "%s%s = [", base_indent, name, level=level, stacklevel=stacklevel)

    for i, item in enumerate(items, start=1):
        if max_items and i > max_items and level >= logging.DEBUG:
            _log(
                logger,
                "%s# ... %d more items of type %r",
                nested_indent,
                len(items) - max_items,
                type(item),
                level=level,
                stacklevel=stacklevel,
            )
            _log(
                logger,
                "%s# change level to 'DEBUG' to print all values",
                nested_indent,
                level=level,
                stacklevel=stacklevel,
            )
            break
        _log(logger, "%s%r,", nested_indent, item, level=level, stacklevel=stacklevel)

    _log(logger, "%s]", base_indent, level=level, stacklevel=stacklevel)


def entity_boundary_log(logger: logging.Logger, msg: str, char: str = "=", stacklevel: int = 1) -> None:
    """Prints message with boundary characters.

    Examples
    --------

    .. code:: python

        entity_boundary_log(logger, "Begin")
        entity_boundary_log(logger, "End", "-")

    .. code-block:: text

        =================== Begin ====================
        ------------------- End ----------------------

    """
    filing = char * (HALF_SCREEN_SIZE - len(msg) // 2)
    _log(logger, "%s %s %s", filing, msg, filing, stacklevel=stacklevel + 1)


def log_options(
    logger: logging.Logger,
    options: dict | None,
    name: str = "options",
    indent: int = 0,
    stacklevel: int = 1,
    **kwargs,
):
    """Log options dict in following format:

    Examples
    --------

    .. code:: python

        log_options(
            logger,
            Options(some="value", abc=1, bcd=None, cde=True, feg=SomeEnum.VALUE).dict(
                by_alias=True
            ),
        )
        log_options(logger, None)

    .. code-block:: text

        INFO  onetl.module        options = {
        INFO  onetl.module            'some': 'value',
        INFO  onetl.module            'abc': 1,
        INFO  onetl.module            'bcd': None,
        INFO  onetl.module            'cde': True,
        INFO  onetl.module            'feg': SomeEnum.VALUE,
        INFO  onetl.module        }

        INFO  onetl.module        options = None

    """

    stacklevel += 1
    if options:
        log_with_indent(logger, "%s = {", name, indent=indent, stacklevel=stacklevel, **kwargs)
        for option, value in options.items():
            value_wrapped = f"'{value}'" if isinstance(value, Enum) else repr(value)
            log_with_indent(
                logger,
                "%r: %s,",
                option,
                value_wrapped,
                indent=indent + 4,
                stacklevel=stacklevel,
                **kwargs,
            )
        log_with_indent(logger, "}", indent=indent, stacklevel=stacklevel, **kwargs)
    else:
        log_with_indent(logger, "%s = %r", name, None, indent=indent, stacklevel=stacklevel, **kwargs)


def log_dataframe_schema(logger: logging.Logger, df: DataFrame, indent: int = 0, stacklevel: int = 1):
    """Log dataframe schema in the following format:

    Examples
    --------

    .. code:: python

        log_dataframe_schema(logger, df)

    .. code-block:: text

        root
        |-- age: integer (nullable = true)
        |-- name: string (nullable = true)

    """

    stacklevel += 1
    log_with_indent(logger, "df_schema:", indent=indent, stacklevel=stacklevel)

    schema_tree = io.StringIO()
    with redirect_stdout(schema_tree):
        # unfortunately, printSchema immediately prints tree instead of returning it
        # so we need a hack
        df.printSchema()

    for line in schema_tree.getvalue().splitlines():
        log_with_indent(logger, "%s", line, indent=indent + 4, stacklevel=stacklevel)


def log_hwm(logger: logging.Logger, hwm: HWM, indent: int = 0, stacklevel: int = 1):
    """Log HWM in the following format:

    Examples
    --------

    .. code:: python

        hwm = ColumnIntHWM(name="my_unique_name", source="my_source", value=123)
        log_hwm(logger, hwm)

    .. code-block:: text

        INFO  onetl.module        hwm = ColumnIntHWM(
        INFO  onetl.module            name = "my_unique_name",
        INFO  onetl.module            entity = "my_source",
        INFO  onetl.module            expression = None,
        INFO  onetl.module            value = 123,
        INFO  onetl.module        )

    .. code-block:: text

        INFO  onetl.module        hwm = FileListHWM(
        INFO  onetl.module            name = "my_unique_name",
        INFO  onetl.module            entity = "my_source",
        INFO  onetl.module            expression = None,
        INFO  onetl.module            value = [
        INFO  onetl.module                AbsolutePath("/some/file1.csv"),
        INFO  onetl.module                AbsolutePath("/some/file2.csv"),
        INFO  onetl.module                AbsolutePath("/some/file3.csv"),
        INFO  onetl.module            ]
        INFO  onetl.module        )
    """
    stacklevel += 1

    log_with_indent(logger, "hwm = %s(", type(hwm).__name__, indent=indent, stacklevel=stacklevel)
    log_with_indent(logger, "name = %r,", hwm.name, indent=indent + 4, stacklevel=stacklevel)
    if hwm.description:
        log_with_indent(logger, "description = %r,", hwm.name, indent=indent + 4, stacklevel=stacklevel)
    log_with_indent(logger, "entity = %r,", hwm.entity, indent=indent + 4, stacklevel=stacklevel)
    log_with_indent(logger, "expression = %r,", hwm.expression, indent=indent + 4, stacklevel=stacklevel)
    if hwm.value is not None:
        if isinstance(hwm.value, Collection):
            log_collection(logger, "value", hwm.value, max_items=10, indent=indent + 4, stacklevel=stacklevel)
        else:
            log_with_indent(logger, "value = %r,", hwm.value, indent=indent + 4, stacklevel=stacklevel)

    log_with_indent(logger, ")", indent=indent, stacklevel=stacklevel)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import Enum
from typing import Iterable, Optional, Tuple

from ordered_set import OrderedSet

try:
    from pydantic.v1 import PrivateAttr, validator
except (ImportError, AttributeError):
    from pydantic import PrivateAttr, validator  # type: ignore[no-redef, assignment]

from onetl._util.file import generate_temp_path
from onetl.base import BaseFileConnection
from onetl.base.path_protocol import PathWithStatsProtocol
from onetl.base.pure_path_protocol import PurePathProtocol
from onetl.exception import DirectoryNotFoundError, NotAFileError
from onetl.file.file_set import FileSet
from onetl.file.file_uploader.options import FileUploaderOptions
from onetl.file.file_uploader.result import UploadResult
from onetl.hooks import slot, support_hooks
from onetl.impl import (
    FailedLocalFile,
    FileExistBehavior,
    FrozenModel,
    LocalPath,
    RemotePath,
    path_repr,
)
from onetl.log import entity_boundary_log, log_lines, log_options, log_with_indent

log = logging.getLogger(__name__)

# source, target, temp
UPLOAD_ITEMS_TYPE = OrderedSet[Tuple[LocalPath, RemotePath, Optional[RemotePath]]]


class FileUploadStatus(Enum):
    SUCCESSFUL = 0
    FAILED = 1
    SKIPPED = 2
    MISSING = -1


@support_hooks
class FileUploader(FrozenModel):
    """Allows you to upload files to a remote source with specified file connection
    and parameters, and return an object with upload result summary. |support_hooks|

    .. note::

        This class is used to upload files **only** from local directory to the remote one.

        It does NOT support direct file transfer between filesystems, like ``FTP -> SFTP``.
        You should use :ref:`file-downloader` + FileUploader to implement ``FTP -> local dir -> SFTP``.

    .. warning::

        This class does **not** support read strategies.

    .. versionadded:: 0.1.0

    .. versionchanged:: 0.8.0
        Moved ``onetl.core.FileDownloader`` → ``onetl.file.FileDownloader``

    Parameters
    ----------
    connection : :obj:`onetl.connection.FileConnection`
        Class which contains File system connection properties. See :ref:`file-connections` section.

    target_path : os.PathLike or str
        Remote path where want you upload files to

    local_path : os.PathLike or str, optional, default: ``None``
        The local directory from which the data is loaded.

        Could be ``None``, but only if you pass absolute file paths directly to
        :obj:`~run` method

        .. versionadded:: 0.3.0

    temp_path : os.PathLike or str, optional, default: ``None``
        If set, this path will be used for uploading a file, and then renaming it to the target file path.
        If ``None`` (default since v0.5.0) is passed, files are uploaded directly to ``target_path``.

        .. warning::

            In case of production ETL pipelines, please set a value for ``temp_path`` (NOT ``None``).
            This allows to properly handle upload interruption,
            without creating half-uploaded files in the target,
            because unlike file upload, ``rename`` call is atomic.

        .. warning::

            In case of connections like SFTP or FTP, which can have multiple underlying filesystems,
            please pass ``temp_path`` path on the SAME filesystem as ``target_path``.
            Otherwise instead of ``rename``, remote OS will move file between filesystems,
            which is NOT atomic operation.

        .. versionchanged:: 0.5.0
            Default changed from ``/tmp`` to ``None``

    options : :obj:`~FileUploader.Options` | dict | None, default: ``None``
        File upload options. See :obj:`FileUploader.Options <onetl.file.file_uploader.options.FileUploaderOptions>`

    Examples
    --------
    Simple Uploader creation

    .. code:: python

        from onetl.connection import HDFS
        from onetl.file import FileUploader

        hdfs = HDFS(...)

        uploader = FileUploader(
            connection=hdfs,
            target_path="/path/to/remote/source",
        )

    Uploader with all parameters

    .. code:: python

        from onetl.connection import HDFS
        from onetl.file import FileUploader

        hdfs = HDFS(...)

        uploader = FileUploader(
            connection=hdfs,
            target_path="/path/to/remote/source",
            temp_path="/user/onetl",
            local_path="/some/local/directory",
            options=FileUploader.Options(delete_local=True, if_exists="overwrite"),
        )

    """

    Options = FileUploaderOptions

    connection: BaseFileConnection

    target_path: RemotePath

    local_path: Optional[LocalPath] = None
    temp_path: Optional[RemotePath] = None

    options: FileUploaderOptions = FileUploaderOptions()

    _connection_checked: bool = PrivateAttr(default=False)

    @slot
    def run(self, files: Iterable[str | os.PathLike] | None = None) -> UploadResult:
        """
        Method for uploading files to remote host. |support_hooks|

        .. versionadded:: 0.1.0

        Parameters
        ----------

        files : Iterator[str | os.PathLike] | None, default ``None``
            File list to upload.

            If empty, upload files from ``local_path``.

        Returns
        -------
        :obj:`UploadResult <onetl.file.file_uploader.upload_result.UploadResult>`

            Upload result object

        Raises
        ------
        :obj:`onetl.exception.DirectoryNotFoundError`

            ``local_path`` does not found

        NotADirectoryError

            ``local_path`` is not a directory

        ValueError

            File in ``files`` argument does not match ``local_path``

        Examples
        --------

        Upload files from ``local_path`` to ``target_path``:

        >>> from onetl.file import FileUploader
        >>> uploader = FileUploader(local_path="/local", target_path="/remote", ...)
        >>> upload_result = uploader.run()
        >>> upload_result
        UploadResult(
            successful=FileSet([
                RemoteFile("/remote/file1"),
                RemoteFile("/remote/file2"),
                # directory structure is preserved
                RemoteFile("/remote/nested/path/file3")
            ]),
            failed=FileSet([
                FailedLocalFile("/local/failed.file"),
            ]),
            skipped=FileSet([
                LocalPath("/local/already.exists"),
            ]),
            missing=FileSet([
                LocalPath("/local/missing.file"),
            ]),
        )

        Upload only certain files from ``local_path``:

        >>> from onetl.file import FileUploader
        >>> uploader = FileUploader(local_path="/local", target_path="/remote", ...)
        >>> # paths could be relative or absolute, but all should be in "/local"
        >>> upload_result = uploader.run(
        ...     [
        ...         "/local/file1",
        ...         "/local/nested/path/file3",
        ...         # excluding "/local/file2",
        ...     ]
        ... )
        >>> upload_result
        UploadResult(
            successful=FileSet([
                RemoteFile("/remote/file1"),
                # directory structure is preserved
                RemoteFile("/remote/nested/path/file3"),
            ]),
            failed=FileSet([]),
            skipped=FileSet([]),
            missing=FileSet([]),
        )

        Upload only certain files from any folder:

        >>> from onetl.file import FileUploader
        >>> uploader = FileUploader(target_path="/remote", ...)  # no local_path set
        >>> # only absolute paths
        >>> upload_result = uploader.run(
        ...     [
        ...         "/local/file1.txt",
        ...         "/any/nested/path/file3.txt",
        ...     ]
        ... )
        >>> upload_result
        UploadResult(
            successful=FileSet([
                RemoteFile("/remote/file1.txt"),
                # directory structure is NOT preserved without local_path
                RemoteFile("/remote/file3.txt"),
            ]),
            failed=FileSet([]),
            skipped=FileSet([]),
            missing=FileSet([]),
        )
        """

        entity_boundary_log(log, f"{self.__class__.__name__}.run() starts")

        if files is None and not self.local_path:
            raise ValueError("Neither file list nor `local_path` are passed")

        if not self._connection_checked:
            self._log_parameters(files)

            if self.local_path:
                self._check_local_path()
            self.connection.check()
            self._connection_checked = True

        self.connection.create_dir(self.target_path)

        if files is None:
            log.debug("|%s| File list is not passed to `run` method", self.__class__.__name__)
            files = self.view_files()

        if not files:
            log.info("|%s| No files to upload!", self.__class__.__name__)
            return UploadResult()

        current_temp_dir: RemotePath | None = None
        if self.temp_path:
            current_temp_dir = generate_temp_path(self.temp_path)

        to_upload = self._validate_files(files, current_temp_dir=current_temp_dir)

        # remove folder only after everything is checked
        if self.options.if_exists == FileExistBehavior.REPLACE_ENTIRE_DIRECTORY:
            self.connection.remove_dir(self.target_path, recursive=True)
            self.connection.create_dir(self.target_path)

        if current_temp_dir:
            current_temp_dir = self.connection.create_dir(current_temp_dir)

        result = self._upload_files(to_upload)

        if current_temp_dir:
            self._remove_temp_dir(current_temp_dir)

        self._log_result(result)
        entity_boundary_log(log, f"{self.__class__.__name__}.run() ends", char="-")
        return result

    @slot
    def view_files(self) -> FileSet[LocalPath]:
        """
        Get file list in the ``local_path``. |support_hooks|

        .. versionadded:: 0.3.0

        Raises
        ------
        :obj:`onetl.exception.DirectoryNotFoundError`

            ``local_path`` does not found

        NotADirectoryError

            ``local_path`` is not a directory

        Returns
        -------
        FileSet[LocalPath]
            Set of files in ``local_path``

        Examples
        --------

        View files:

        >>> from onetl.file import FileUploader
        >>> uploader = FileUploader(local_path="/local", ...)
        >>> uploader.view_files()
        FileSet([
            LocalPath("/local/file1.txt"),
            LocalPath("/local/file3.txt"),
            LocalPath("/local/nested/path/file3.txt"),
        ])
        """

        if not self.local_path:
            raise ValueError("Cannot call `.view_files()` without `local_path`")

        log.debug("|Local FS| Getting files list from path '%s'", self.local_path)

        if not self._connection_checked:
            self._check_local_path()

        result = FileSet()

        try:
            for root, dirs, files in os.walk(self.local_path):
                log.debug("|Local FS| Listing dir '%s': %d dirs, %d files", root, len(dirs), len(files))
                result.update(LocalPath(root) / file for file in files)
        except Exception as e:
            raise RuntimeError(
                f"Couldn't read directory tree from local dir '{self.local_path}'",
            ) from e

        return result

    @validator("local_path", pre=True, always=True)
    def _resolve_local_path(cls, local_path):
        return LocalPath(local_path).resolve() if local_path else None

    @validator("target_path", pre=True, always=True)
    def _validate_target_path(cls, target_path):
        return RemotePath(target_path)

    @validator("temp_path", pre=True, always=True)
    def _validate_temp_path(cls, temp_path):
        return RemotePath(temp_path) if temp_path else None

    def _log_parameters(self, files: Iterable[str | os.PathLike] | None = None) -> None:
        log.info("|Local FS| -> |%s| Uploading files using parameters:'", self.connection.__class__.__name__)
        log_with_indent(log, "local_path = %s", f"'{self.local_path}'" if self.local_path else "None")
        log_with_indent(log, "target_path = '%s'", self.target_path)
        log_with_indent(log, "temp_path = %s", f"'{self.temp_path}'" if self.temp_path else "None")
        log_options(log, self.options.dict(by_alias=True))

        if self.options.delete_local:
            log.warning("|%s| LOCAL FILES WILL BE PERMANENTLY DELETED AFTER UPLOADING !!!", self.__class__.__name__)

        if self.options.if_exists == FileExistBehavior.REPLACE_ENTIRE_DIRECTORY:
            log.warning("|%s| TARGET DIRECTORY WILL BE CLEANED UP BEFORE UPLOADING FILES !!!", self.__class__.__name__)

        if files and self.local_path:
            log.warning(
                "|%s| Passed both `local_path` and files list at the same time. Using explicit files list",
                self.__class__.__name__,
            )

    def _validate_files(  # noqa: WPS231
        self,
        local_files: Iterable[os.PathLike | str],
        current_temp_dir: RemotePath | None,
    ) -> UPLOAD_ITEMS_TYPE:
        result = OrderedSet()

        for file in local_files:
            local_file_path = LocalPath(file)
            local_file = local_file_path
            tmp_file: RemotePath | None = None

            if not self.local_path:
                # Upload into a flat structure
                if not local_file_path.is_absolute():
                    raise ValueError("Cannot pass relative file path with empty `local_path`")

                filename = local_file_path.name
                target_file = self.target_path / filename
                if current_temp_dir:
                    tmp_file = current_temp_dir / filename  # noqa: WPS220
            else:
                # Upload according to source folder structure
                if self.local_path in local_file_path.parents:
                    # Make relative remote path
                    target_file = self.target_path / local_file_path.relative_to(self.local_path)
                    if current_temp_dir:
                        tmp_file = current_temp_dir / local_file_path.relative_to(self.local_path)  # noqa: WPS220
                elif not local_file_path.is_absolute():
                    # Passed path is already relative
                    local_file = self.local_path / local_file_path
                    target_file = self.target_path / local_file_path
                    if current_temp_dir:
                        tmp_file = current_temp_dir / local_file_path  # noqa: WPS220
                else:
                    # Wrong path (not relative path and source path not in the path to the file)
                    raise ValueError(f"File path '{local_file}' does not match source_path '{self.local_path}'")

            if local_file.exists() and not local_file.is_file():
                raise NotAFileError(f"{path_repr(local_file)} is not a file")

            result.add((local_file, target_file, tmp_file))

        return result

    def _check_local_path(self):
        if not self.local_path.exists():
            raise DirectoryNotFoundError(f"'{self.local_path}' does not exist")

        if not self.local_path.is_dir():
            raise NotADirectoryError(f"{path_repr(self.local_path)} is not a directory")

    def _upload_files(self, to_upload: UPLOAD_ITEMS_TYPE) -> UploadResult:
        files = FileSet(item[0] for item in to_upload)
        log.info("|%s| Files to be uploaded:", self.__class__.__name__)
        log_lines(log, str(files))
        log_with_indent(log, "")
        log.info("|%s| Starting the upload process...", self.__class__.__name__)

        self._create_dirs(to_upload)

        result = UploadResult()
        for status, file in self._bulk_upload(to_upload):
            if status == FileUploadStatus.SUCCESSFUL:
                result.successful.add(file)
            elif status == FileUploadStatus.FAILED:
                result.failed.add(file)
            elif status == FileUploadStatus.SKIPPED:
                result.skipped.add(file)
            elif status == FileUploadStatus.MISSING:
                result.missing.add(file)

        return result

    def _create_dirs(
        self,
        to_upload: UPLOAD_ITEMS_TYPE,
    ) -> None:
        """
        Create all parent paths before uploading files
        This is required to avoid errors then multiple threads create the same dir
        """
        parent_paths = OrderedSet()
        for _, target_file, tmp_file in to_upload:
            parent_paths.add(target_file.parent)
            if tmp_file:
                parent_paths.add(tmp_file.parent)

        for parent_path in parent_paths:
            self.connection.create_dir(parent_path)

    def _bulk_upload(
        self,
        to_upload: UPLOAD_ITEMS_TYPE,
    ) -> list[tuple[FileUploadStatus, PurePathProtocol | PathWithStatsProtocol]]:
        workers = self.options.workers
        files_count = len(to_upload)
        result = []

        real_workers = workers
        if files_count < workers:
            log.debug(
                "|%s| Asked for %d workers, but there are only %d files",
                self.__class__.__name__,
                workers,
                files_count,
            )
            real_workers = files_count

        if real_workers > 1:
            log.debug("|%s| Using ThreadPoolExecutor with %d workers", self.__class__.__name__, real_workers)
            with ThreadPoolExecutor(
                max_workers=workers,
                thread_name_prefix=self.__class__.__name__,
            ) as executor:
                futures = [
                    executor.submit(self._upload_file, local_file, target_file, tmp_file)
                    for local_file, target_file, tmp_file in to_upload
                ]
                for future in as_completed(futures):
                    result.append(future.result())
        else:
            log.debug("|%s| Using plain old for-loop", self.__class__.__name__)
            for local_file, target_file, tmp_file in to_upload:
                result.append(
                    self._upload_file(
                        local_file,
                        target_file,
                        tmp_file,
                    ),
                )

        return result

    def _upload_file(  # noqa: WPS231
        self,
        local_file: LocalPath,
        target_file: RemotePath,
        tmp_file: RemotePath | None,
    ) -> tuple[FileUploadStatus, PurePathProtocol | PathWithStatsProtocol]:
        if tmp_file:
            log.info(
                "|%s| Uploading file '%s' to '%s' (via tmp '%s')",
                self.__class__.__name__,
                local_file,
                target_file,
                tmp_file,
            )
        else:
            log.info("|%s| Uploading file '%s' to '%s'", self.__class__.__name__, local_file, target_file)

        if not local_file.exists():
            log.warning("|%s| Missing file '%s', skipping", self.__class__.__name__, local_file)
            return FileUploadStatus.MISSING, local_file

        try:
            replace = False
            if self.connection.path_exists(target_file):
                file = self.connection.resolve_file(target_file)
                if self.options.if_exists == FileExistBehavior.ERROR:
                    raise FileExistsError(f"File {path_repr(file)} already exists")

                if self.options.if_exists == FileExistBehavior.IGNORE:
                    log.warning("|%s| File %s already exists, skipping", self.__class__.__name__, path_repr(file))
                    return FileUploadStatus.SKIPPED, local_file

                replace = True

            if tmp_file:
                # Files are loaded to temporary directory before moving them to target dir.
                # This prevents operations with partly uploaded files

                self.connection.upload_file(local_file, tmp_file)
                uploaded_file = self.connection.rename_file(tmp_file, target_file, replace=replace)
            else:
                # Direct upload
                uploaded_file = self.connection.upload_file(local_file, target_file, replace=replace)

            if self.options.delete_local:
                local_file.unlink()
                log.warning("|Local FS| Successfully removed file %s", local_file)

            return FileUploadStatus.SUCCESSFUL, uploaded_file

        except Exception as e:
            if log.isEnabledFor(logging.DEBUG):
                log.exception("|%s| Couldn't upload file to target dir", self.__class__.__name__, exc_info=e)
            else:
                log.exception("|%s| Couldn't upload file to target dir: %s", self.__class__.__name__, e, exc_info=False)

            return FileUploadStatus.FAILED, FailedLocalFile(path=local_file, exception=e)

    def _remove_temp_dir(self, temp_dir: RemotePath) -> None:
        try:
            self.connection.remove_dir(temp_dir, recursive=True)
        except Exception:
            log.exception("|%s| Error while removing temp directory", self.__class__.__name__)

    def _log_result(self, result: UploadResult) -> None:
        log.info("")
        log.info("|%s| Upload result:", self.__class__.__name__)
        log_lines(log, str(result))



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import Any, Optional

from etl_entities.hwm import HWM

from onetl.impl import BaseModel
from onetl.strategy.batch_hwm_strategy import BatchHWMStrategy
from onetl.strategy.hwm_strategy import HWMStrategy


class OffsetMixin(BaseModel):
    hwm: Optional[HWM] = None
    offset: Any = None

    def fetch_hwm(self) -> None:
        super().fetch_hwm()

        if self.hwm and self.hwm.value is not None and self.offset is not None:
            self.hwm -= self.offset


class IncrementalStrategy(OffsetMixin, HWMStrategy):
    """Incremental strategy for :ref:`db-reader`/:ref:`file-downloader`.

    Used for fetching only new rows/files from a source
    by filtering items not covered by the previous :ref:`HWM` value.

    For :ref:`db-reader`:
        First incremental run is just the same as :obj:`SnapshotStrategy <onetl.strategy.snapshot_strategy.SnapshotStrategy>`:

        .. code:: sql

            SELECT id, data FROM mydata;

        Then the max value of ``id`` column (e.g. ``1000``) will be saved as ``HWM`` to :ref:`HWM Store <hwm>`.

        Next incremental run will read only new data from the source:

        .. code:: sql

            SELECT id, data FROM mydata WHERE id > 1000; -- hwm value

        Pay attention to resulting dataframe **does not include** row with ``id=1000`` because it has been read before.

        .. warning::

            If code inside the context manager raised an exception, like:

            .. code:: python

                with IncrementalStrategy():
                    df = reader.run()  # something went wrong here
                    writer.run(df)  # or here
                    # or here...

            When DBReader will **NOT** update HWM in HWM Store.
            This allows to resume reading process from the *last successful run*.

    For :ref:`file-downloader`:
        Behavior depends on ``hwm`` type.

        ``hwm=FileListHWM(...)``:
            First incremental run is just the same as :obj:`SnapshotStrategy <onetl.strategy.snapshot_strategy.SnapshotStrategy>` -
            all files are downloaded:

            .. code:: bash

                $ hdfs dfs -ls /path

                /path/my/file1
                /path/my/file2

            .. code:: python

                DownloadResult(
                    ...,
                    successful={
                        LocalFile("/downloaded/file1"),
                        LocalFile("/downloaded/file2"),
                    },
                )

            Then the downloaded files list is saved as ``FileListHWM`` object into :ref:`HWM Store <hwm>`:

            .. code:: python

                FileListHWM(
                    ...,
                    entity="/path",
                    value=[
                        "/path/my/file1",
                        "/path/my/file2",
                    ],
                )

            Next incremental run will download only new files from the source:

            .. code:: bash

                $ hdfs dfs -ls /path

                /path/my/file1
                /path/my/file2
                /path/my/file3

            .. code:: python

                # only files which are not in FileListHWM
                DownloadResult(
                    ...,
                    successful={
                        LocalFile("/downloaded/file3"),
                    },
                )

            New files will be added to the ``FileListHWM`` and saved to :ref:`HWM Store <hwm>`:

            .. code:: python

                FileListHWM(
                    ...,
                    entity="/path",
                    value=[
                        "/path/my/file1",
                        "/path/my/file2",
                        "/path/my/file3",
                    ],
                )

        .. warning::

            FileDownload updates HWM in HWM Store at the end of ``.run()`` call,
            **NOT** while exiting strategy context. This is because:

            * FileDownloader does not raise exceptions if some file cannot be downloaded.
            * FileDownloader creates files on local filesystem, and file content may differ for different :obj:`modes <onetl.file.file_downloader.file_downloader.FileDownloader.Options.mode>`.
            * It can remove files from the source if :obj:`delete_source <onetl.file.file_downloader.file_downloader.FileDownloader.Options.delete_source>` is set to ``True``.

    .. versionadded:: 0.1.0

    Parameters
    ----------
    offset : Any, default: ``None``

        If passed, the offset value will be used to read rows which appeared in the source after the previous read.

        For example, previous incremental run returned rows:

        .. code::

            898
            899
            900
            1000

        Current HWM value is 1000.

        But since then few more rows appeared in the source:

        .. code::

            898
            899
            900
            901 # new
            902 # new
            ...
            999 # new
            1000

        and you need to read them too.

        So you can set ``offset=100``, so a next incremental run will generate SQL query like:

        .. code:: sql

            SELECT id, data FROM public.mydata WHERE id > 900;
            -- 900 = 1000 - 100 = hwm - offset

        and return rows since 901 (**not** 900), **including** 1000 which was already captured by HWM.

        .. warning::

            This can lead to reading duplicated values from the table.
            You probably need additional deduplication step to handle them

        .. warning::

            Cannot be used with :ref:`file-downloader` and ``hwm=FileListHWM(...)``

        .. note::

            ``offset`` value will be subtracted from the HWM, so it should have a proper type.

            For example, for ``TIMESTAMP`` column ``offset`` type should be :obj:`datetime.timedelta`, not :obj:`int`

    Examples
    --------

    Incremental run with :ref:`db-reader`:

    .. code:: python

        from onetl.connection import Postgres
        from onetl.db import DBReader
        from onetl.strategy import IncrementalStrategy

        from pyspark.sql import SparkSession

        maven_packages = Postgres.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        postgres = Postgres(
            host="postgres.domain.com",
            user="myuser",
            password="*****",
            database="target_database",
            spark=spark,
        )

        reader = DBReader(
            connection=postgres,
            source="public.mydata",
            columns=["id", "data"],
            hwm=DBReader.AutoDetectHWM(name="some_hwm_name", expression="id"),
        )

        writer = DBWriter(connection=hive, target="newtable")

        with IncrementalStrategy():
            df = reader.run()
            writer.run(df)

    .. code:: sql

        -- previous HWM value was 1000
        -- DBReader will generate query like:

        SELECT id, data
        FROM public.mydata
        WHERE id > 1000; --- from HWM (EXCLUDING first row)

    Incremental run with :ref:`db-reader` and ``offset``:

    .. code:: python

        ...

        with IncrementalStrategy(offset=100):
            df = reader.run()
            writer.run(df)

    .. code:: sql

        -- previous HWM value was 1000
        -- DBReader will generate query like:

        SELECT id, data
        FROM public.mydata
        WHERE id > 900; -- from HWM-offset (EXCLUDING first row)

    ``hwm.expression`` can be a date or datetime, not only integer:

    .. code:: python

        from datetime import timedelta

        reader = DBReader(
            connection=postgres,
            source="public.mydata",
            columns=["business_dt", "data"],
            hwm=DBReader.AutoDetectHWM(name="some_hwm_name", expression="business_dt"),
        )

        with IncrementalStrategy(offset=timedelta(days=1)):
            df = reader.run()
            writer.run(df)

    .. code:: sql

        -- previous HWM value was '2021-01-10'
        -- DBReader will generate query like:

        SELECT business_dt, data
        FROM public.mydata
        WHERE business_dt > CAST('2021-01-09' AS DATE); -- from HWM-offset (EXCLUDING first row)

    Incremental run with :ref:`db-reader` and :ref:`kafka` connection
    (by ``offset`` in topic - :etl-entities:`KeyValueHWM <hwm/key_value/index.html>`):

    .. code:: python

        from onetl.connection import Kafka
        from onetl.db import DBReader
        from onetl.strategy import IncrementalStrategy

        from pyspark.sql import SparkSession

        maven_packages = Kafka.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        kafka = Kafka(
            addresses=["mybroker:9092", "anotherbroker:9092"],
            cluster="my-cluster",
            spark=spark,
        )

        reader = DBReader(
            connection=kafka,
            source="topic_name",
            hwm=DBReader.AutoDetectHWM(name="some_hwm_name", expression="offset"),
        )

        with IncrementalStrategy():
            df = reader.run()

    Incremental run with :ref:`file-downloader` and ``hwm=FileListHWM(...)``:

    .. code:: python

        from onetl.connection import SFTP
        from onetl.file import FileDownloader
        from onetl.strategy import SnapshotStrategy
        from etl_entities.hwm import FileListHWM

        sftp = SFTP(
            host="sftp.domain.com",
            user="user",
            password="*****",
        )

        downloader = FileDownloader(
            connection=sftp,
            source_path="/remote",
            local_path="/local",
            hwm=FileListHWM(name="some_hwm_name"),
        )

        with IncrementalStrategy():
            df = downloader.run()

        # current run will download only files which were not downloaded in previous runs
    """


class IncrementalBatchStrategy(OffsetMixin, BatchHWMStrategy):
    """Incremental batch strategy for :ref:`db-reader`.

    .. note::

        Cannot be used with :ref:`file-downloader`

    Same as :obj:`IncrementalStrategy <onetl.strategy.incremental_strategy.IncrementalStrategy>`,
    but reads data from the source in sequential batches (1..N) like:

    .. code:: sql

        1: SELECT id, data
           FROM public.mydata
           WHERE id > 1000 AND id <= 1100; -- previous HWM value is 1000, step is 100

        2: WHERE id > 1100 AND id <= 1200; -- + step
        3: WHERE id > 1200 AND id <= 1200; -- + step
        N: WHERE id > 1300 AND id <= 1400; -- until stop

    This allows to use less CPU and RAM than reading all the data in the one batch,
    but takes proportionally more time.

    .. warning::

        Unlike :obj:`SnapshotBatchStrategy <onetl.strategy.snapshot_strategy.SnapshotBatchStrategy>`,
        it **saves** current HWM value after **each batch** into :ref:`HWM Store <hwm>`.

        So if code inside the context manager raised an exception, like:

        .. code:: python

            with IncrementalBatchStrategy() as batches:
                for _ in batches:
                    df = reader.run()  # something went wrong here
                    writer.run(df)  # or here
                    # or here...

        DBReader will **NOT** update HWM in HWM Store for the failed batch.

        All of that allows to resume reading process from the *last successful batch*.

    .. warning::

        Not every :ref:`DB connection <db-connections>`
        supports batch strategy. For example, Kafka connection doesn't support it.
        Make sure the connection you use is compatible with the IncrementalBatchStrategy.

    .. versionadded:: 0.1.0

    Parameters
    ----------
    step : Any

        Step size used for generating batch SQL queries like:

        .. code:: sql

            SELECT id, data
            FROM public.mydata
            WHERE id > 1000 AND id <= 1100; -- 1000 is previous HWM value, step is 100

        .. note::

            Step defines a range of values will be fetched by each batch. This is **not**
            a number of rows, it depends on a table content and value distribution across the rows.

        .. note::

            ``step`` value will be added to the HWM, so it should have a proper type.

            For example, for ``TIMESTAMP`` column ``step`` type should be :obj:`datetime.timedelta`, not :obj:`int`

    stop : Any, default: ``None``

        If passed, the value will be used for generating WHERE clauses with ``hwm.expression`` filter,
        as a stop value for the last batch.

        If not set, the value is determined by a separated query:

        .. code:: sql

            SELECT MAX(id) as stop
            FROM public.mydata
            WHERE id > 1000; -- 1000 is previous HWM value (if any)

        .. note::

            ``stop`` should be the same type as ``hwm.expression`` value,
            e.g. :obj:`datetime.datetime` for ``TIMESTAMP`` column, :obj:`datetime.date` for ``DATE``, and so on

    offset : Any, default: ``None``

        If passed, the offset value will be used to read rows which appeared in the source after the previous read.

        For example, previous incremental run returned rows:

        .. code::

            898
            899
            900
            1000

        Current HWM value is 1000.

        But since then few more rows appeared in the source:

        .. code::

            898
            899
            900
            901 # new
            902 # new
            ...
            999 # new
            1000

        and you need to read them too.

        So you can set ``offset=100``, so the first batch of a next incremental run will look like:

        .. code:: sql

            SELECT id, data
            FROM public.mydata
            WHERE id > 900 AND id <= 1000; -- 900 = 1000 - 100 = HWM - offset

        and return rows from 901 (**not** 900) to **1000** (duplicate).

        .. warning::

            This can lead to reading duplicated values from the table.
            You probably need additional deduplication step to handle them

        .. note::

            ``offset`` value will be subtracted from the HWM, so it should have a proper type.

            For example, for ``TIMESTAMP`` column ``offset`` type should be :obj:`datetime.timedelta`, not :obj:`int`

    Examples
    --------

    IncrementalBatch run:

    .. code:: python

        from onetl.connection import Postgres, Hive
        from onetl.db import DBReader
        from onetl.strategy import IncrementalBatchStrategy

        from pyspark.sql import SparkSession

        maven_packages = Postgres.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        postgres = Postgres(
            host="postgres.domain.com",
            user="myuser",
            password="*****",
            database="target_database",
            spark=spark,
        )

        hive = Hive(cluster="rnd-dwh", spark=spark)

        reader = DBReader(
            connection=postgres,
            source="public.mydata",
            columns=["id", "data"],
            hwm=DBReader.AutoDetectHWM(name="some_hwm_name", expression="id"),
        )

        writer = DBWriter(connection=hive, target="newtable")

        with IncrementalBatchStrategy(step=100) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- previous HWM value was 1000
        -- each batch (1..N) will perform a query which return some part of input data

        1:  SELECT id, data
            FROM public.mydata
            WHERE id > 1100 AND id <= 1200; --- from HWM to HWM+step (EXCLUDING first row)

        2:  WHERE id > 1200 AND id <= 1300; -- + step
        N:  WHERE id > 1300 AND id <= 1400; -- until max value of HWM column

    IncrementalBatch run with ``stop`` value:

    .. code:: python

        with IncrementalBatchStrategy(step=100, stop=2000) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- previous HWM value was 1000
        -- each batch (1..N) will perform a query which return some part of input data

        1:  SELECT id, data
            FROM public.mydata
            WHERE id > 1000 AND id <= 1100; --- from HWM to HWM+step (EXCLUDING first row)

        2:  WHERE id > 1100 AND id <= 1200; -- + step
        ...
        N:  WHERE id > 1900 AND id <= 2000; -- until stop

    IncrementalBatch run with ``offset`` value:

    .. code:: python

        with IncrementalBatchStrategy(step=100, offset=100) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- previous HWM value was 1000
        -- each batch (1..N) will perform a query which return some part of input data

        1:  SELECT id, data
            FROM public.mydata
            WHERE id >  900 AND id <= 1000; --- from HWM-offset to HWM-offset+step (EXCLUDING first row)

        2:  WHERE id > 1000 AND id <= 1100; -- + step
        3:  WHERE id > 1100 AND id <= 1200; -- + step
        ...
        N:  WHERE id > 1300 AND id <= 1400; -- until max value of HWM column

    IncrementalBatch run with all possible options:

    .. code:: python

        with IncrementalBatchStrategy(
            step=100,
            stop=2000,
            offset=100,
        ) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- previous HWM value was 1000
        -- each batch (1..N) will perform a query which return some part of input data

        1:  SELECT id, data
            FROM public.mydata
            WHERE id > 900 AND id <= 1000; --- from HWM-offset to HWM-offset+step (EXCLUDING first row)

        2:  WHERE id > 1000 AND id <= 1100; -- + step
        3:  WHERE id > 1100 AND id <= 1200; -- + step
        ...
        N:  WHERE id > 1900 AND id <= 2000; -- until stop

    ``hwm.expression`` can be a date or datetime, not only integer:

    .. code:: python

        from datetime import date, timedelta

        reader = DBReader(
            connection=postgres,
            source="public.mydata",
            columns=["business_dt", "data"],
            hwm=DBReader.AutoDetectHWM(name="some_hwm_name", expression="business_dt"),
        )

        with IncrementalBatchStrategy(
            step=timedelta(days=5),
            stop=date("2021-01-31"),
            offset=timedelta(days=1),
        ) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- previous HWM value was '2021-01-10'
        -- each batch (1..N) will perform a query which return some part of input data

        1:  SELECT business_dt, data
            FROM public.mydata
            WHERE business_dt  > CAST('2021-01-09' AS DATE)  -- from HWM-offset (EXCLUDING first row)
            AND   business_dt <= CAST('2021-01-14' AS DATE); -- to HWM-offset+step

        2:  WHERE business_dt  > CAST('2021-01-14' AS DATE) -- + step
            AND   business_dt <= CAST('2021-01-19' AS DATE);

        3:  WHERE business_dt  > CAST('2021-01-19' AS DATE) -- + step
            AND   business_dt <= CAST('2021-01-24' AS DATE);

        ...

        N:  WHERE business_dt  > CAST('2021-01-29' AS DATE)
            AND   business_dt <= CAST('2021-01-31' AS DATE); -- until stop

    """

    def __next__(self):
        self.save_hwm()
        return super().__next__()

    @classmethod
    def _log_exclude_fields(cls) -> set[str]:
        return super()._log_exclude_fields() | {"start"}



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import os
import textwrap
from typing import Generic, TypeVar

from humanize import naturalsize
from ordered_set import OrderedSet

from onetl.base import PathProtocol, PathWithStatsProtocol, PurePathProtocol
from onetl.exception import EmptyFilesError, ZeroFileSizeError
from onetl.impl import path_repr

T = TypeVar("T", bound=PurePathProtocol)
INDENT = " " * 4


class FileSet(OrderedSet[T], Generic[T]):
    """
    Ordered set of pathlib-like objects.

    It has all the methods of generic set (e.g. ``add``, ``difference``, ``intersection``),
    as well as list (e.g. ``append``, ``index``, ``[]``).

    It also has a ``total_size`` helper method.
    """

    @property
    def total_size(self) -> int:
        """
        Get total size (in bytes) of files in the set

        Examples
        --------

        >>> from onetl.impl import LocalPath
        >>> from onet.file.file_set import FileSet
        >>> file_set = FileSet({LocalPath("/some/file"), LocalPath("/some/another.file")})
        >>> path_set.total_size  # in bytes
        1024
        """

        return sum(
            file.stat().st_size or 0 for file in self if isinstance(file, PathWithStatsProtocol) and file.exists()
        )

    def raise_if_empty(self) -> None:
        """
        Raise exception if there are no files in the set

        Raises
        ------
        EmptyFilesError

            File set is empty

        Examples
        --------

        >>> from onet.file.file_set import FileSet
        >>> file_set = FileSet()
        >>> file_set.raise_if_empty()
        Traceback (most recent call last):
            ...
        onetl.exception.EmptyFilesError: There are no files in the set
        """

        if not self:
            raise EmptyFilesError("There are no files in the set")

    def raise_if_contains_zero_size(self) -> None:
        """
        Raise exception if file set contains a file with zero size

        Raises
        ------
        ZeroFileSizeError

            File set contains a file with zero size

        Examples
        --------

        >>> from onetl.impl import RemoteFile, LocalPath
        >>> from onet.file.file_set import FileSet
        >>> file_set = FileSet(
        ...     LocalPath("/local/empty1.file"),
        ...     LocalPath("/local/empty2.file"),
        ...     LocalPath("/local/normal.file"),
        ... )
        >>> file_set.raise_if_contains_zero_size()
        Traceback (most recent call last):
            ...
        onetl.exception.ZeroFileSizeError: 2 files out of 3 have zero size:
            '/local/empty1.file'
            '/local/empty2.file'
        """

        lines = []
        for file in self:
            if isinstance(file, PathProtocol) and not file.exists():
                continue

            if isinstance(file, PathWithStatsProtocol) and file.stat().st_size > 0:
                continue

            lines.append(repr(os.fspath(file)))

        if not lines:
            return

        lines_str = textwrap.indent(os.linesep.join(lines), INDENT)
        file_number_str = f"{len(lines)} files" if len(lines) > 1 else "1 file"
        error_message = f"{file_number_str} out of {len(self)} have zero size:{os.linesep}{lines_str}"

        raise ZeroFileSizeError(error_message)

    @property
    def summary(self) -> str:
        """
        Return summary about files in the set

        Examples
        --------

        >>> from onetl.impl import LocalPath
        >>> from onet.file.file_set import FileSet
        >>> path_set1 = FileSet(
        ...     [
        ...         LocalPath("/local/file"),
        ...         LocalPath("/local/another.file"),
        ...     ]
        ... )
        >>> print(path_set1.summary)
        2 files (30.7 kB)
        >>> path_set2 = FileSet()
        >>> print(path_set2.summary)
        No files
        """

        if not self:
            return "No files"

        file_number_str = f"{len(self)} files" if len(self) > 1 else "1 file"
        return f"{file_number_str} (size='{naturalsize(self.total_size)}')"

    @property
    def details(self) -> str:  # noqa: WPS473
        """
        Return detailed information about files in the set

        Examples
        --------

        >>> from onetl.impl import LocalPath
        >>> from onet.file.file_set import FileSet
        >>> path_set1 = FileSet(
        ...     [
        ...         LocalPath("/local/file"),
        ...         LocalPath("/local/another.file"),
        ...     ]
        ... )
        >>> print(path_set1.details)
        2 files (30.7 kB):
            '/local/file' (10.2 kB)
            '/local/another.file' (20.5 kB)

        >>> path_set2 = FileSet()
        >>> print(path_set2.details)
        No files
        """

        if not self:
            return self.summary

        lines = [path_repr(file, with_mode=False, with_kind=False, with_owner=False, with_mtime=False) for file in self]

        summary = f"{self.summary}:{os.linesep}{INDENT}"

        lines_str = textwrap.indent(os.linesep.join(lines), INDENT).strip()
        return summary + lines_str

    def __str__(self) -> str:
        """Same as :obj:`onetl.file.file_set.file_set.FileSet.details`"""
        return self.details



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from enum import Enum

try:
    from pydantic.v1 import Field, root_validator
except (ImportError, AttributeError):
    from pydantic import Field, root_validator  # type: ignore[no-redef, assignment]

from onetl.impl import GenericOptions

PROHIBITED_OPTIONS = frozenset(
    (
        "assign",
        "endingOffsets",
        "endingOffsetsByTimestamp",
        "endingTimestamp",
        "kafka.*",
        "startingOffsets",
        "startingOffsetsByTimestamp",
        "startingTimestamp",
        "subscribe",
        "subscribePattern",
        "topic",
    ),
)

KNOWN_READ_OPTIONS = frozenset(
    (
        "failOnDataLoss",
        "fetchOffset.numRetries",
        "fetchOffset.retryIntervalMs",
        "groupIdPrefix",
        "kafkaConsumer.pollTimeoutMs",
        "maxOffsetsPerTrigger",
        "maxTriggerDelay",
        "minOffsetsPerTrigger",
        "minPartitions",
    ),
)


class KafkaTopicExistBehaviorKafka(str, Enum):
    ERROR = "error"
    APPEND = "append"

    def __str__(self) -> str:
        return str(self.value)


class KafkaReadOptions(GenericOptions):
    """Reading options for Kafka connector.

    .. note ::

        You can pass any value
        `supported by connector <https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html>`_,
        even if it is not mentioned in this documentation.

        The set of supported options depends on connector version. See link above.

    .. warning::

        Options:
            * ``assign``
            * ``endingOffsets``
            * ``endingOffsetsByTimestamp``
            * ``kafka.*``
            * ``startingOffsets``
            * ``startingOffsetsByTimestamp``
            * ``startingTimestamp``
            * ``subscribe``
            * ``subscribePattern``

        are populated from connection attributes, and cannot be overridden by the user in ``ReadOptions`` to avoid issues.

    .. versionadded:: 0.9.0

    Examples
    --------

    Read options initialization

    .. code:: python

        options = Kafka.ReadOptions(
            include_headers=False,
            minPartitions=50,
        )
    """

    include_headers: bool = Field(default=False, alias="includeHeaders")
    """
    If ``True``, add ``headers`` column to output DataFrame.

    If ``False``, column will not be added.
    """

    class Config:
        prohibited_options = PROHIBITED_OPTIONS
        known_options = KNOWN_READ_OPTIONS
        extra = "allow"


class KafkaWriteOptions(GenericOptions):
    """Writing options for Kafka connector.

    .. note ::

        You can pass any value
        `supported by connector <https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html>`_,
        even if it is not mentioned in this documentation.

        The set of supported options depends on connector version. See link above.

    .. warning::

        Options:
            * ``kafka.*``
            * ``topic``

        are populated from connection attributes, and cannot be overridden by the user in ``WriteOptions`` to avoid issues.

    .. versionadded:: 0.9.0

    Examples
    --------

    Write options initialization

    .. code:: python

        options = Kafka.WriteOptions(
            if_exists="append",
            include_headers=True,
        )
    """

    if_exists: KafkaTopicExistBehaviorKafka = Field(default=KafkaTopicExistBehaviorKafka.APPEND)
    """Behavior of writing data into existing topic.

    Same as ``df.write.mode(...)``.

    Possible values:
        * ``append`` (default) - Adds new objects into existing topic.
        * ``error`` - Raises an error if topic already exists.
    """

    include_headers: bool = Field(default=False, alias="includeHeaders")
    """
    If ``True``, ``headers`` column from dataframe can be written to Kafka (requires Kafka 2.0+).

    If ``False`` and dataframe contains ``headers`` column, an exception will be raised.
    """

    class Config:
        prohibited_options = PROHIBITED_OPTIONS | KNOWN_READ_OPTIONS
        known_options: frozenset[str] = frozenset()
        extra = "allow"

    @root_validator(pre=True)
    def _mode_is_restricted(cls, values):
        if "mode" in values:
            raise ValueError("Parameter `mode` is not allowed. Please use `if_exists` parameter instead.")
        return values



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, ClassVar, Dict, Optional

try:
    from pydantic.v1 import Field, validator
except (ImportError, AttributeError):
    from pydantic import Field, validator  # type: ignore[no-redef, assignment]

from onetl._util.java import try_import_java_class
from onetl._util.scala import get_default_scala_version
from onetl._util.spark import get_spark_version
from onetl._util.version import Version
from onetl.exception import MISSING_JVM_CLASS_MSG
from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrameReader, DataFrameWriter, SparkSession


PROHIBITED_OPTIONS = frozenset(
    (
        # These options should be passed in Spark session config, not file format options
        "spark.*",
    ),
)

READ_WRITE_OPTIONS = frozenset(
    ("positionalFieldMatching",),
)

READ_OPTIONS = frozenset(
    (
        "ignoreExtension",
        "datetimeRebaseMode",
    ),
)

WRITE_OPTIONS = frozenset(
    (
        "recordName",
        "recordNamespace",
        "compression",
    ),
)

log = logging.getLogger(__name__)


@support_hooks
class Avro(ReadWriteFileFormat):
    """
    Avro file format. |support_hooks|

    Based on `Spark Avro <https://spark.apache.org/docs/latest/sql-data-sources-avro.html>`_ file format.

    Supports reading/writing files with ``.avro`` extension.

    .. dropdown:: Version compatibility

        * Spark versions: 2.4.x - 3.5.x
        * Java versions: 8 - 20

        See documentation from link above.

    .. note ::

        You can pass any option to the constructor, even if it is not mentioned in this documentation.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version. See link above.

    .. versionadded:: 0.9.0

    Examples
    --------

    Describe options how to read from/write to Avro file with specific options:

    .. code:: python

        from onetl.file.format import Avro
        from pyspark.sql import SparkSession

        # Create Spark session with Avro package loaded
        maven_packages = Avro.get_packages(spark_version="3.5.2")
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        # Describe file format
        schema = {
            "type": "record",
            "name": "Person",
            "fields": [
                {"name": "name", "type": "string"},
                {"name": "age", "type": "int"},
            ],
        }
        avro = Avro(schema_dict=schema, compression="snappy")

    """

    name: ClassVar[str] = "avro"

    schema_dict: Optional[Dict] = Field(default=None, alias="avroSchema")
    schema_url: Optional[str] = Field(default=None, alias="avroSchemaUrl")

    class Config:
        known_options = READ_WRITE_OPTIONS | READ_OPTIONS | WRITE_OPTIONS
        prohibited_options = PROHIBITED_OPTIONS
        extra = "allow"

    @slot
    @classmethod
    def get_packages(
        cls,
        spark_version: str,
        scala_version: str | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        See `Maven package index <https://mvnrepository.com/artifact/org.apache.spark/spark-avro>`_
        for all available packages.

        .. versionadded:: 0.9.0

        Parameters
        ----------
        spark_version : str
            Spark version in format ``major.minor.patch``.

        scala_version : str, optional
            Scala version in format ``major.minor``.

            If ``None``, ``spark_version`` is used to determine Scala version.

        Examples
        --------

        .. code:: python

            from onetl.file.format import Avro

            Avro.get_packages(spark_version="3.2.4")
            Avro.get_packages(spark_version="3.2.4", scala_version="2.12")

        """

        spark_ver = Version(spark_version).min_digits(3)
        if spark_ver < Version("2.4"):
            raise ValueError(f"Spark version should be at least 2.4, got {spark_version}")

        scala_ver = Version(scala_version).min_digits(2) if scala_version else get_default_scala_version(spark_ver)
        if scala_ver < Version("2.11"):
            raise ValueError(f"Scala version should be at least 2.11, got {scala_ver.format('{0}.{1}')}")

        return [f"org.apache.spark:spark-avro_{scala_ver.format('{0}.{1}')}:{spark_ver.format('{0}.{1}.{2}')}"]

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        java_class = "org.apache.spark.sql.avro.AvroFileFormat"

        try:
            try_import_java_class(spark, java_class)
        except Exception as e:
            spark_version = get_spark_version(spark)
            msg = MISSING_JVM_CLASS_MSG.format(
                java_class=java_class,
                package_source=self.__class__.__name__,
                args=f"spark_version='{spark_version}'",
            )
            if log.isEnabledFor(logging.DEBUG):
                log.debug("Missing Java class", exc_info=e, stack_info=True)
            raise ValueError(msg) from e

    @slot
    def apply_to_reader(self, reader: DataFrameReader) -> DataFrameReader:
        options = self.dict(by_alias=True, exclude_none=True, exclude={"schema"})
        if self.schema_dict:
            options["avroSchema"] = json.dumps(self.schema_dict)
        return reader.format(self.name).options(**options)

    @slot
    def apply_to_writer(self, writer: DataFrameWriter) -> DataFrameWriter:
        options = self.dict(by_alias=True, exclude_none=True, exclude={"schema"})
        if self.schema_dict:
            options["avroSchema"] = json.dumps(self.schema_dict)
        return writer.format(self.name).options(**options)

    def parse_column(self, column: str | Column) -> Column:
        """
        Parses an Avro binary column into a structured Spark SQL column using Spark's
        `from_avro <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.avro.functions.from_avro.html>`_ function, based on the schema provided within the class.

        .. note::

            Can be used only with Spark 3.x+

        .. warning::

            If ``schema_url`` is provided, ``requests`` library is used to fetch the schema from the URL. It should be installed manually, like this:

            .. code:: bash

                pip install requests

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing Avro bytes to deserialize.
            Schema should match the provided Avro schema.

        Returns
        -------
        Column with deserialized data. Schema is matching the provided Avro schema. Column name is the same as input column.

        Raises
        ------
        ValueError
            If the Spark version is less than 3.x or if neither schema_dict nor schema_url is defined.
        ImportError
            If ``schema_url`` is used and the ``requests`` library is not installed.

        Examples
        --------

        >>> from pyspark.sql.functions import decode
        >>> from onetl.file.format import Avro
        >>> df.show()
        +----+----------------------+----------+---------+------+-----------------------+-------------+
        |key |value                 |topic     |partition|offset|timestamp              |timestampType|
        +----+----------------------+----------+---------+------+-----------------------+-------------+
        |[31]|[0A 41 6C 69 63 65 28]|topicAvro |0        |0     |2024-04-24 13:02:25.911|0            |
        |[32]|[06 42 6F 62 32]      |topicAvro |0        |1     |2024-04-24 13:02:25.922|0            |
        +----+----------------------+----------+---------+------+-----------------------+-------------+
        >>> df.printSchema()
        root
        |-- key: binary (nullable = true)
        |-- value: binary (nullable = true)
        |-- topic: string (nullable = true)
        |-- partition: integer (nullable = true)
        |-- offset: integer (nullable = true)
        |-- timestamp: timestamp (nullable = true)
        |-- timestampType: integer (nullable = true)
        >>> avro = Avro(
        ...     schema_dict={
        ...         "type": "record",
        ...         "name": "Person",
        ...         "fields": [
        ...             {"name": "name", "type": "string"},
        ...             {"name": "age", "type": "int"},
        ...         ],
        ...     }
        ... )
        >>> parsed_df = df.select(decode("key", "UTF-8").alias("key"), avro.parse_column("value"))
        >>> parsed_df.show(truncate=False)
        +---+-----------+
        |key|value      |
        +---+-----------+
        |1  |{Alice, 20}|
        |2  |{Bob, 25}  |
        +---+-----------+
        >>> parsed_df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        """
        from pyspark.sql import Column, SparkSession  # noqa: WPS442
        from pyspark.sql.functions import col

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_spark_version_for_serialization(spark)

        from pyspark.sql.avro.functions import from_avro

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa: WPS437
        else:
            column_name, column = column, col(column).cast("binary")

        schema = self._get_schema_json()
        if not schema:
            raise ValueError("Avro.parse_column can be used only with defined `schema_dict` or `schema_url`")

        return from_avro(column, schema).alias(column_name)

    def serialize_column(self, column: str | Column) -> Column:
        """
        Serializes a structured Spark SQL column into an Avro binary column using Spark's
        `to_avro <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.avro.functions.to_avro.html#pyspark.sql.avro.functions.to_avro>`_ function.

        .. note::

            Can be used only with Spark 3.x+

        .. warning::

            If ``schema_url`` is provided, ``requests`` library is used to fetch the schema from the URL. It should be installed manually, like this:

            .. code:: bash

                pip install requests

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing the data to serialize to Avro format.

        Returns
        -------
        Column with binary Avro data. Column name is the same as input column.

        Raises
        ------
        ValueError
            If the Spark version is less than 3.x.
        ImportError
            If ``schema_url`` is used and the ``requests`` library is not installed.

        Examples
        --------

        >>> from pyspark.sql.functions import decode
        >>> from onetl.file.format import Avro
        >>> df.show()
        +---+-----------+
        |key|value      |
        +---+-----------+
        |1  |{Alice, 20}|
        |2  |  {Bob, 25}|
        +---+-----------+
        >>> df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        >>> # serializing data into Avro format
        >>> avro = Avro(
        ...     schema_dict={
        ...         "type": "record",
        ...         "name": "Person",
        ...         "fields": [
        ...             {"name": "name", "type": "string"},
        ...             {"name": "age", "type": "int"},
        ...         ],
        ...     }
        ... )
        >>> serialized_df = df.select("key", avro.serialize_column("value"))
        >>> serialized_df.show(truncate=False)
        +---+----------------------+
        |key|value                 |
        +---+----------------------+
        |  1|[0A 41 6C 69 63 65 28]|
        |  2|[06 42 6F 62 32]      |
        +---+----------------------+
        >>> serialized_df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: binary (nullable = true)
        """
        from pyspark.sql import Column, SparkSession  # noqa: WPS442
        from pyspark.sql.functions import col

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_spark_version_for_serialization(spark)

        from pyspark.sql.avro.functions import to_avro

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa:  WPS437
        else:
            column_name, column = column, col(column)

        schema = self._get_schema_json()
        return to_avro(column, schema).alias(column_name)

    @validator("schema_dict", pre=True)
    def _parse_schema_from_json(cls, value):
        if isinstance(value, (str, bytes)):
            return json.loads(value)
        return value

    def _check_spark_version_for_serialization(self, spark: SparkSession):
        spark_version = get_spark_version(spark)
        if spark_version.major < 3:
            class_name = self.__class__.__name__
            error_msg = (
                f"`{class_name}.parse_column` or `{class_name}.serialize_column` are available "
                f"only since Spark 3.x, but got {spark_version}."
            )
            raise ValueError(error_msg)

    def _get_schema_json(self) -> str:
        if self.schema_dict:
            return json.dumps(self.schema_dict)
        elif self.schema_url:
            try:
                import requests

                response = requests.get(self.schema_url)  # noqa: S113
                return response.text
            except ImportError as e:
                raise ImportError(
                    "The 'requests' library is required to use 'schema_url' but is not installed. "
                    "Install it with 'pip install requests' or avoid using 'schema_url'.",
                ) from e
        else:
            return ""



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from datetime import date, datetime

from onetl.connection.db_connection.jdbc_connection import JDBCDialect


class TeradataDialect(JDBCDialect):
    # https://docs.teradata.com/r/w4DJnG9u9GdDlXzsTXyItA/lkaegQT4wAakj~K_ZmW1Dg
    def get_partition_column_hash(self, partition_column: str, num_partitions: int) -> str:
        return f"HASHAMP(HASHBUCKET(HASHROW({partition_column}))) mod {num_partitions}"

    def get_partition_column_mod(self, partition_column: str, num_partitions: int) -> str:
        return f"{partition_column} mod {num_partitions}"

    def _serialize_datetime(self, value: datetime) -> str:
        result = value.isoformat()
        return f"CAST('{result}' AS TIMESTAMP)"

    def _serialize_date(self, value: date) -> str:
        result = value.isoformat()
        return f"CAST('{result}' AS DATE)"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


@dataclass
class Edge:
    value: Any = None
    including: bool = True

    def is_set(self) -> bool:
        return self.value is not None


@dataclass
class Window:
    expression: str
    start_from: Edge = field(default_factory=Edge)
    stop_at: Edge = field(default_factory=Edge)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class PathStatProtocol(Protocol):
    """
    Protocol for ``os.stat_result``-like objects.

    Includes only minimal set of fields supported by any file system
    """

    @property
    def st_size(self) -> int:
        """
        Size of file, in bytes
        """

    @property
    def st_mtime(self) -> float | None:
        """
        Unix timestamp of most recent content modification
        """

    @property
    def st_uid(self) -> str | int | None:
        """
        User identifier of the file owner
        """

    @property
    def st_gid(self) -> str | int | None:
        """
        Group identifier of the file owner
        """

    @property
    def st_mode(self) -> int | None:
        """
        File mode bits
        """



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import sys
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from functools import wraps
from typing import Callable, Generator, Generic, TypeVar

from typing_extensions import Protocol, runtime_checkable

from onetl.log import NOTICE

logger = logging.getLogger(__name__)

T = TypeVar("T")


class HookPriority(int, Enum):
    """
    Hook priority enum.

    All hooks within the same priority are executed in the same order they were registered.

    .. versionadded:: 0.7.0
    """

    FIRST = -1
    "Hooks with this priority will run first."

    NORMAL = 0
    "Hooks with this priority will run after :obj:`~FIRST` but before :obj:`~LAST`."

    LAST = 1
    "Hooks with this priority will run last."


@dataclass  # noqa: WPS338
class Hook(Generic[T]):  # noqa: WPS338
    """
    Hook representation.

    .. versionadded:: 0.7.0

    Parameters
    ----------

        callback : :obj:`typing.Callable`

            Some callable object which will be wrapped into a Hook, like function or ContextManager class.

        enabled : bool

            Will hook be executed or not. Useful for debugging.

        priority : :obj:`onetl.hooks.hook.HookPriority`

            Changes hooks priority, see ``HookPriority`` documentation.

    Examples
    --------

    .. code:: python

        from onetl.hooks.hook import Hook, HookPriority


        def some_func(*args, **kwargs): ...


        hook = Hook(callback=some_func, enabled=True, priority=HookPriority.FIRST)
    """

    callback: Callable[..., T]
    enabled: bool = True
    priority: HookPriority = HookPriority.NORMAL

    def __post_init__(self):
        self.priority = HookPriority(self.priority)

    def enable(self):
        """
        Enable the hook.

        .. versionadded:: 0.7.0

        Examples
        --------

        >>> def func1(): ...
        >>> hook = Hook(callback=func1, enabled=False)
        >>> hook.enabled
        False
        >>> hook.enable()
        >>> hook.enabled
        True
        """
        if self.enabled:
            logger.log(
                NOTICE,
                "|Hooks| Hook '%s.%s' already enabled",
                self.callback.__module__,
                self.callback.__qualname__,
            )
        else:
            logger.log(NOTICE, "|Hooks| Enable hook '%s.%s'", self.callback.__module__, self.callback.__qualname__)
            self.enabled = True

    def disable(self):
        """
        Disable the hook.

        .. versionadded:: 0.7.0

        Examples
        --------

        >>> def func1(): ...
        >>> hook = Hook(callback=func1, enabled=True)
        >>> hook.enabled
        True
        >>> hook.disable()
        >>> hook.enabled
        False
        """
        if self.enabled:
            logger.log(NOTICE, "|Hooks| Disable hook '%s.%s'", self.callback.__module__, self.callback.__qualname__)
            self.enabled = False
        else:
            logger.log(
                NOTICE,
                "|Hooks| Hook '%s.%s' already disabled",
                self.callback.__module__,
                self.callback.__qualname__,
            )

    @contextmanager
    def skip(self):
        """
        Temporary disable the hook.

        .. note::

            If hook was created with ``enabled=False``, or was disabled by :obj:`~disable`,
            its state will left intact after exiting the context.

            You should call :obj:`~enable` explicitly to change its state.

        .. versionadded:: 0.7.0

        Examples
        --------

        .. tabs::

            .. tab:: Context manager syntax

                >>> def func1(): ...
                >>> hook = Hook(callback=func1, enabled=True)
                >>> hook.enabled
                True
                >>> with hook.skip():
                ...     print(hook.enabled)
                False
                >>> # hook state is restored as it was before entering the context manager
                >>> hook.enabled
                True

            .. tab:: Decorator syntax

                >>> def func1(): ...
                >>> hook = Hook(callback=func1, enabled=True)
                >>> hook.enabled
                True
                >>> @hook.skip()
                ... def hook_disabled():
                ...     print(hook.enabled)
                >>> hook_disabled()
                False
                >>> # hook state is restored as it was before entering the context manager
                >>> hook.enabled
                True
        """
        if not self.enabled:
            logger.log(
                NOTICE,
                "|Hooks| Hook '%s.%s' already disabled, nothing to skip",
                self.callback.__module__,
                self.callback.__qualname__,
            )
            yield
        else:
            logger.log(
                NOTICE,
                "|Hooks| Skipping hook '%s.%s' ...",
                self.callback.__module__,
                self.callback.__qualname__,
            )
            self.enabled = False

            yield

            logger.log(
                NOTICE,
                "|Hooks| Restoring hook '%s.%s' ...",
                self.callback.__module__,
                self.callback.__qualname__,
            )
            self.enabled = True

    def __call__(self, *args, **kwargs) -> T | ContextDecorator:
        """
        Calls the original callback with passed args.

        Examples
        --------

        >>> from onetl.hooks.hook import Hook, HookPriority
        >>> def some_func(*args, **kwargs):
        ...     print(args)
        ...     print(kwargs)
        ...     return "func result"
        >>> hook = Hook(callback=some_func)
        >>> result = hook(1, "abc", some="arg")
        (1, 'abc')
        {'some': 'arg'}
        >>> result
        'func result'
        """
        result = self.callback(*args, **kwargs)
        if isinstance(result, Generator):
            return ContextDecorator(result)
        return result


@runtime_checkable
class CanProcessResult(Protocol):
    """
    Protocol which should be implemented by ContextManager
    allowing it to process result of original method call and modify/replace it with something else.
    """

    def process_result(self, result: T) -> T: ...


class ContextDecorator:
    """
    Helper for :obj:`~hook` decorator.

    Analog of :obj:`contextlib._GeneratorContextManager`, but instead of generator function
    plus arguments it accepts generator object.

    Also it does not implement ``__call__``, so it does not allow for context manager to be
    used as a decorator.
    """

    def __init__(self, gen: Generator):
        self.gen: Generator = gen
        self.first_yield_result = None

    def __enter__(self):
        """
        Start generator and stop at first ``yield``.

        If generator instead of:

        .. code:: python

            result = yield
            ...
            yield result

        looks like:

        .. code:: python

            yield something

        Just remember this output and return it in :obj:`~process_result` as is.
        """

        try:
            self.first_yield_result = self.gen.send(None)
        except StopIteration:
            pass

        return self

    def __exit__(self, exc_type, value, traceback):  # noqa: WPS231
        """
        Copy of :obj:`contextlib._GeneratorContextManager.__exit__`
        """

        if exc_type is None:
            try:
                next(self.gen)
            except StopIteration:
                return False
            raise RuntimeError("generator didn't stop")

        if value is None:
            # Need to force instantiation so we can reliably
            # tell if we get the same exception back
            value = value or exc_type()

        try:
            self.gen.throw(exc_type, value, traceback)
        except StopIteration as exc:
            # Suppress StopIteration *unless* it's the same exception that
            # was passed to throw().  This prevents a StopIteration
            # raised inside the "with" statement from being suppressed.
            return exc is not value
        except RuntimeError as exc:
            # Don't re-raise the passed in exception. (issue27122)
            if exc is value:
                return False
            # Likewise, avoid suppressing if a StopIteration exception
            # was passed to throw() and later wrapped into a RuntimeError
            # (see PEP 479).
            if exc_type is StopIteration and exc.__cause__ is value:
                return False
            raise
        except:  # noqa: E722, B001
            # only re-raise if it's *not* the exception that was
            # passed to throw(), because __exit__() must not raise
            # an exception unless __exit__() itself failed.  But throw()
            # has to raise the exception to signal propagation, so this
            # fixes the impedance mismatch between the throw() protocol
            # and the __exit__() protocol.
            #
            # This cannot use 'except BaseException as exc' (as in the
            # async implementation) to maintain compatibility with
            # Python 2, where old-style class exceptions are not caught
            # by 'except BaseException'.
            if sys.exc_info()[1] is value:
                return False
            raise
        raise RuntimeError("generator didn't stop after throw()")

    def process_result(self, result):
        """
        Handle original method call result, and return new value.

        If generator instead of:

        .. code:: python

            result = yield
            # this is there ``process_result`` is called
            yield result

        looks like:

        .. code:: python

            yield something

        Just return the yielded result.
        """

        try:
            generator_result = self.gen.send(result)
            if generator_result is not None:
                return generator_result
        except StopIteration:
            return self.first_yield_result

        return None


def hook(inp: Callable[..., T] | None = None, enabled: bool = True, priority: HookPriority = HookPriority.NORMAL):
    """
    Initialize hook from callable/context manager.

    .. versionadded:: 0.7.0

    Examples
    --------

    .. tabs::

        .. code-tab:: py Decorate a function or generator

            from onetl.hooks import hook, HookPriority


            @hook
            def some_func(*args, **kwargs):
                ...


            @hook(enabled=True, priority=HookPriority.FIRST)
            def another_func(*args, **kwargs):
                ...

        .. code-tab:: py Decorate a context manager

            from onetl.hooks import hook, HookPriority


            @hook
            class SimpleContextManager:
                def __init__(self, *args, **kwargs):
                    ...

                def __enter__(self):
                    ...
                    return self

                def __exit__(self, exc_type, exc_value, traceback):
                    ...
                    return False


            @hook(enabled=True, priority=HookPriority.FIRST)
            class ContextManagerWithProcessResult:
                def __init__(self, *args, **kwargs):
                    ...

                def __enter__(self):
                    ...
                    return self

                def __exit__(self, exc_type, exc_value, traceback):
                    ...
                    return False

                def process_result(self, result):
                    # special method to handle method result call
                    return modify(result)

                ...
    """

    def inner_wrapper(callback: Callable[..., T]):  # noqa: WPS430
        if isinstance(callback, Hook):
            raise TypeError("@hook decorator can be applied only once")

        result = Hook(callback=callback, enabled=enabled, priority=priority)
        return wraps(callback)(result)

    if inp is None:
        return inner_wrapper

    return inner_wrapper(inp)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from datetime import date, datetime

from onetl.connection.db_connection.jdbc_connection import JDBCDialect


class MySQLDialect(JDBCDialect):
    def get_partition_column_hash(self, partition_column: str, num_partitions: int) -> str:
        return f"MOD(CONV(CONV(RIGHT(MD5({partition_column}), 16), 16, 2), 2, 10), {num_partitions})"

    def get_partition_column_mod(self, partition_column: str, num_partitions: int) -> str:
        return f"MOD({partition_column}, {num_partitions})"

    def escape_column(self, value: str) -> str:
        return f"`{value}`"

    def _serialize_datetime(self, value: datetime) -> str:
        result = value.strftime("%Y-%m-%d %H:%M:%S.%f")
        return f"STR_TO_DATE('{result}', '%Y-%m-%d %H:%i:%s.%f')"

    def _serialize_date(self, value: date) -> str:
        result = value.strftime("%Y-%m-%d")
        return f"STR_TO_DATE('{result}', '%Y-%m-%d')"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except (ImportError, AttributeError):
    from pydantic import Field, PrivateAttr, validator  # type: ignore[no-redef, assignment]

from onetl._metrics.command import SparkCommandMetrics
from onetl._metrics.recorder import SparkMetricsRecorder
from onetl._util.spark import override_job_description
from onetl.base import BaseDBConnection
from onetl.hooks import slot, support_hooks
from onetl.impl import FrozenModel, GenericOptions
from onetl.log import (
    entity_boundary_log,
    log_dataframe_schema,
    log_lines,
    log_options,
    log_with_indent,
)

if TYPE_CHECKING:
    from pyspark.sql import DataFrame

log = logging.getLogger(__name__)


@support_hooks
class DBWriter(FrozenModel):
    """Class specifies schema and table where you can write your dataframe. |support_hooks|

    .. versionadded:: 0.1.0

    .. versionchanged:: 0.8.0
        Moved ``onetl.core.DBReader`` → ``onetl.db.DBReader``

    Parameters
    ----------
    connection : :obj:`onetl.connection.DBConnection`
        Class which contains DB connection properties. See :ref:`db-connections` section.

    target : str
        Table/collection/etc name to write data to.

        If connection has schema support, you need to specify the full name of the source
        including the schema, e.g. ``schema.name``.

        .. versionchanged:: 0.7.0
            Renamed ``table`` → ``target``

    options : dict, :obj:`onetl.connection.DBConnection.WriteOptions`, default: ``None``
        Spark write options. Can be in form of special ``WriteOptions`` object or a dict.

        For example:
        ``{"if_exists": "replace_entire_table", "compression": "snappy"}``
        or
        ``Hive.WriteOptions(if_exists="replace_entire_table", compression="snappy")``

        .. note::

            Some sources does not support writing options.


    Examples
    --------
    Simple Writer creation:

    .. code:: python

        from onetl.connection import Postgres
        from onetl.db import DBWriter
        from pyspark.sql import SparkSession

        maven_packages = Postgres.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        postgres = Postgres(
            host="postgres.domain.com",
            user="your_user",
            password="***",
            database="target_db",
            spark=spark,
        )

        writer = DBWriter(
            connection=postgres,
            target="fiddle.dummy",
        )

    Writer creation with options:

    .. code:: python

        from onetl.connection import Postgres
        from onetl.db import DBWriter
        from pyspark.sql import SparkSession

        maven_packages = Postgres.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        postgres = Postgres(
            host="postgres.domain.com",
            user="your_user",
            password="***",
            database="target_db",
            spark=spark,
        )

        options = Postgres.WriteOptions(if_exists="replace_entire_table", batchsize=1000)

        writer = DBWriter(
            connection=postgres,
            target="fiddle.dummy",
            options=options,
        )

    Writer to Hive with options:

    .. code:: python

        from onetl.db import DBWriter
        from onetl.connection import Hive
        from pyspark.sql import SparkSession

        spark = SparkSession.builder.appName("spark-app-name").enableHiveSupport().getOrCreate()

        hive = Hive(cluster="rnd-dwh", spark=spark)

        options = {"compression": "snappy", "partitionBy": "id"}
        # or (it is the same):
        options = Hive.WriteOptions(compression="snappy", partitionBy="id")

        writer = DBWriter(
            connection=hive,
            target="default.test",
            options=options,
        )
    """

    connection: BaseDBConnection
    target: str = Field(alias="table")
    options: Optional[GenericOptions] = None

    _connection_checked: bool = PrivateAttr(default=False)

    @validator("target", pre=True, always=True)
    def validate_target(cls, target, values):
        connection: BaseDBConnection = values["connection"]
        return connection.dialect.validate_name(target)

    @validator("options", pre=True, always=True)
    def validate_options(cls, options, values):
        connection = values.get("connection")
        write_options_class = getattr(connection, "WriteOptions", None)
        if write_options_class:
            return write_options_class.parse(options)

        if options:
            raise ValueError(
                f"{connection.__class__.__name__} does not implement WriteOptions, but {options!r} is passed",
            )

        return None

    @slot
    def run(self, df: DataFrame) -> None:
        """
        Method for writing your df to specified target. |support_hooks|

        .. note :: Method does support only **batching** DataFrames.

        .. versionadded:: 0.1.0

        Parameters
        ----------
        df : pyspark.sql.dataframe.DataFrame
            Spark dataframe

        Examples
        --------

        Write dataframe to target:

        .. code:: python

            writer.run(df)
        """
        if df.isStreaming:
            raise ValueError(f"DataFrame is streaming. {self.__class__.__name__} supports only batch DataFrames.")

        entity_boundary_log(log, msg=f"{self.__class__.__name__}.run() starts")

        job_description = f"{self.__class__.__name__}.run({self.target}) -> {self.connection}"
        with override_job_description(self.connection.spark, job_description):
            if not self._connection_checked:
                self._log_parameters()
                log_dataframe_schema(log, df)
                self.connection.check()
                self._connection_checked = True

        with SparkMetricsRecorder(self.connection.spark) as recorder:
            try:
                with override_job_description(self.connection.spark, job_description):
                    self.connection.write_df_to_target(
                        df=df,
                        target=str(self.target),
                        **self._get_write_kwargs(),
                    )
            except Exception:
                metrics = recorder.metrics()
                # SparkListener is not a reliable source of information, metrics may or may not be present.
                # Because of this we also do not return these metrics as method result
                if metrics.output.is_empty:
                    log.error(
                        "|%s| Error while writing dataframe.",
                        self.__class__.__name__,
                    )
                else:
                    log.error(
                        "|%s| Error while writing dataframe. Target MAY contain partially written data!",
                        self.__class__.__name__,
                    )
                self._log_metrics(metrics)
                raise
            finally:
                self._log_metrics(recorder.metrics())

        entity_boundary_log(log, msg=f"{self.__class__.__name__}.run() ends", char="-")

    def _log_parameters(self) -> None:
        log.info("|Spark| -> |%s| Writing DataFrame to target using parameters:", self.connection.__class__.__name__)
        log_with_indent(log, "target = '%s'", self.target)

        options = self.options.dict(by_alias=True, exclude_none=True) if self.options else None
        log_options(log, options)

    def _get_write_kwargs(self) -> dict:
        if self.options:
            return {"options": self.options}

        return {}

    def _log_metrics(self, metrics: SparkCommandMetrics) -> None:
        if not metrics.is_empty:
            log.debug("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
            log_lines(log, str(metrics), level=logging.DEBUG)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import hashlib
import io
import os
from datetime import datetime
from pathlib import Path, PurePath

from onetl.exception import NotAFileError
from onetl.impl import path_repr

# e.g. 20230524122150
DATETIME_FORMAT = "%Y%m%d%H%M%S"


def get_file_hash(
    path: os.PathLike | str,
    algorithm: str,
    chunk_size: int = io.DEFAULT_BUFFER_SIZE,
) -> hashlib._Hash:
    """Get file hash by path and algorithm"""
    digest = hashlib.new(algorithm)
    with open(path, "rb") as file:
        chunk = file.read(chunk_size)
        while chunk:
            digest.update(chunk)
            chunk = file.read(chunk_size)

    return digest


def is_file_readable(path: str | os.PathLike) -> Path:
    """Check if specified path is a file and is readable"""
    path = Path(os.path.expandvars(path)).expanduser().resolve()

    if not path.exists():
        raise FileNotFoundError(f"File '{path}' does not exist")

    if not path.is_file():
        raise NotAFileError(f"{path_repr(path)} is not a file")

    if not os.access(path, os.R_OK):
        raise OSError(f"No read access to file {path_repr(path)}")

    return path


def generate_temp_path(root: PurePath) -> PurePath:
    """
    Returns prefix which will be used for creating temp directory

    Returns
    -------
    RemotePath
        Temp path, containing current host name, process name and datetime

    Examples
    --------

    >>> from etl_entities.process import Process
    >>> from pathlib import Path
    >>> generate_temp_path(Path("/tmp")) # doctest: +SKIP
    Path("/tmp/onetl/currenthost/myprocess/20230524122150")
    >>> with Process(dag="mydag", task="mytask"): # doctest: +SKIP
    ...    generate_temp_path(Path("/abc"))
    Path("/abc/onetl/currenthost/mydag.mytask.myprocess/20230524122150")
    """

    from etl_entities.process import ProcessStackManager

    current_process = ProcessStackManager.get_current()
    current_dt = datetime.now().strftime(DATETIME_FORMAT)
    return root / "onetl" / current_process.host / current_process.full_name / current_dt



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
import textwrap
from contextlib import contextmanager
from math import inf
from typing import TYPE_CHECKING, Any

from onetl._util.version import Version

try:
    from pydantic.v1 import SecretStr
except (ImportError, AttributeError):
    from pydantic import SecretStr  # type: ignore[no-redef, assignment]

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.conf import RuntimeConfig

SPARK_JOB_DESCRIPTION_PROPERTY = "spark.job.description"
SPARK_JOB_GROUP_PROPERTY = "spark.jobGroup.id"


def stringify(value: Any, quote: bool = False) -> Any:  # noqa: WPS212
    """
    Convert values to strings.

    Values ``True``, ``False`` and ``None`` become ``"true"``, ``"false"`` and ``"null"``.

    If input is dict, return dict with stringified values and keys (recursive).

    If input is list, return list with stringified values (recursive).

    If ``quote=True``, wrap string values with double quotes.

    Examples
    --------

    >>> stringify(1)
    '1'
    >>> stringify(True)
    'true'
    >>> stringify(False)
    'false'
    >>> stringify(None)
    'null'
    >>> stringify("string")
    'string'
    >>> stringify("string", quote=True)
    '"string"'
    >>> stringify({"abc": 1})
    {'abc': '1'}
    >>> stringify([1, True, False, None, "string"])
    ['1', 'true', 'false', 'null', 'string']
    """

    if isinstance(value, dict):
        return {stringify(k): stringify(v, quote) for k, v in value.items()}

    if isinstance(value, list):
        return [stringify(v, quote) for v in value]

    if value is None:
        return "null"

    if isinstance(value, bool):
        return "true" if value else "false"

    if isinstance(value, SecretStr):
        value = value.get_secret_value()

    if isinstance(value, os.PathLike):
        value = os.fspath(value)

    if isinstance(value, str):
        return f'"{value}"' if quote else value

    return str(value)


@contextmanager
def inject_spark_param(conf: RuntimeConfig, name: str, value: Any):
    """
    Inject a parameter into a Spark session, and return previous value after exit.

    If value is ``None``, parameter will be reset to default.
    """
    original_value = conf.get(name, None)

    try:
        conf.unset(name)
        if value is not None:
            conf.set(name, value)

        yield
    finally:
        conf.unset(name)
        if original_value is not None:
            conf.set(name, original_value)


def try_import_pyspark():
    """
    Try to import PySpark, or raise exception if missing.
    """
    try:
        import pyspark  # noqa: F401
    except (ImportError, NameError) as e:
        raise ImportError(
            textwrap.dedent(
                """
                Cannot import module "pyspark".

                You should install package as follows:
                    pip install onetl[spark]

                or inject PySpark to sys.path in some other way BEFORE creating class instance.
                """,
            ).strip(),
        ) from e


def get_pyspark_version() -> Version:
    """
    Get Spark version from ``pyspark.__version__``
    """
    try_import_pyspark()
    import pyspark

    return Version(pyspark.__version__)


def get_spark_version(spark_session: SparkSession) -> Version:
    """
    Get Spark version from active Spark session
    """
    return Version(spark_session.version)


def estimate_dataframe_size(spark_session: SparkSession, df: DataFrame) -> int:
    """
    Estimate in-memory DataFrame size in bytes. If cannot be estimated, return 0.

    Using Spark's `SizeEstimator <https://spark.apache.org/docs/3.5.2/api/java/org/apache/spark/util/SizeEstimator.html>`_.
    """
    try:
        size_estimator = spark_session._jvm.org.apache.spark.util.SizeEstimator  # type: ignore[union-attr]
        return size_estimator.estimate(df._jdf)
    except Exception:
        # SizeEstimator uses Java reflection which may behave differently in different Java versions,
        # and also may be prohibited.
        return 0


def get_executor_total_cores(spark_session: SparkSession, include_driver: bool = False) -> tuple[int | float, dict]:
    """
    Calculate maximum number of cores which can be used by Spark on all executors.

    Returns
    -------
    Tuple
        First item is an actual number of cores.
        Second is a dict with config options were used to calculate this value.
    """

    conf = spark_session.sparkContext.getConf()
    config = {}

    master = conf.get("spark.master", "local")

    if "local" in master:
        # no executors, only driver
        expected_cores = spark_session._jvm.Runtime.getRuntime().availableProcessors()  # type: ignore
        config["spark.driver.cores"] = expected_cores
    else:
        cores = int(conf.get("spark.executor.cores", "1"))
        config["spark.executor.cores"] = cores

        dynamic_allocation = conf.get("spark.dynamicAllocation.enabled", "false") == "true"
        if dynamic_allocation:
            # https://spark.apache.org/docs/latest/configuration.html#dynamic-allocation
            # We cannot rely on current executors count because this number depends on
            # current load - it increases while performing heavy calculations, and decreases if executors are idle.
            # If user haven't executed anything in current session, number of executors will be 0.
            #
            # Yes, scheduler can refuse to provide executors == maxExecutors, because
            # queue size limit is reached, or other application has higher priority, so executors were preempted.
            # But cluster health cannot rely on a chance, so pessimistic approach is preferred.

            dynamic_executors = conf.get("spark.dynamicAllocation.maxExecutors", "infinity")
            executors_ratio = int(conf.get("spark.dynamicAllocation.executorAllocationRatio", "1"))
            config["spark.dynamicAllocation.maxExecutors"] = dynamic_executors
            executors = inf if dynamic_executors == "infinity" else int(dynamic_executors * executors_ratio)
        else:
            fixed_executors: float = int(conf.get("spark.executor.instances", "0"))
            config["spark.executor.instances"] = fixed_executors
            executors = fixed_executors

        expected_cores = executors * cores
        if include_driver:
            expected_cores += 1

    return expected_cores, config


@contextmanager
def override_job_description(spark_session: SparkSession, job_description: str):
    """
    Override Spark job description.

    Unlike ``spark_session.sparkContext.setJobDescription``, this method resets job description
    before exiting the context manager, instead of keeping it.

    If user set custom description, it will be left intact.
    """
    spark_context = spark_session.sparkContext
    original_description = spark_context.getLocalProperty(SPARK_JOB_DESCRIPTION_PROPERTY)

    try:
        spark_context.setLocalProperty(SPARK_JOB_DESCRIPTION_PROPERTY, original_description or job_description)
        yield
    finally:
        spark_context.setLocalProperty(SPARK_JOB_DESCRIPTION_PROPERTY, original_description)  # type: ignore[arg-type]
